/**************************************************************************
*   Copyright (C) 2005 by Achal Dhir                                      *
*   achaldhir@gmail.com                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
// opendhcpd.cpp //
#include <string>
#include <math.h>
#include <map>
using namespace std;
#include <sys/types.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <memory.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <syslog.h>
#include "opendhcpd.h"

//Global Variables
timeval tv;
fd_set readfds;
fd_set writefds;
data9 dhcpr;
data9 token;
data1 network;
data1 newNetwork;
data2 cfig;
bool kRunning = true;
dhcpMap dhcpCache;
char tempbuff[512] = "";
char logBuff[256] = "";
char extbuff[256] = "";
MYWORD loggingDay = 0;
bool verbatim = false;
char iniFile[256] = "";
char leaFile[256] = "";
char logFile[256] = "";
char currentLogFile[256] = "";
char arpa[] = ".in-addr.arpa";
time_t t = time(NULL);
pthread_mutex_t mutStateFile = PTHREAD_MUTEX_INITIALIZER;
//pthread_mutex_t mutLogFile = PTHREAD_MUTEX_INITIALIZER;
struct ifconf Ifc;
struct ifreq IfcBuf[MAX_SERVERS];

//constants
char RANGESET[] = "RANGE_SET";
char GLOBALOPTIONS[] = "GLOBAL_OPTIONS";
const char base64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
//const char send200[] = "HTTP/1.1 200 OK\r\nDate: %s\r\nLast-Modified: %s\r\nContent-Type: text/html\r\nConnection: Close\r\nTransfer-Encoding: chunked\r\n";
const char send200[] = "HTTP/1.1 200 OK\r\nDate: %s\r\nLast-Modified: %s\r\nContent-Type: text/html\r\nConnection: Close\r\nContent-Length:         \r\n\r\n";
const char line200[] = "<td>%s</td>";
const char sVersion[] = "Open DHCP Server Version 1.50 Linux Build 1027";
const data4 opData[] =
    {
		{ "SubnetMask", DHCP_OPTION_NETMASK, 3 },
		{ "TimeOffset", DHCP_OPTION_TIMEOFFSET, 4 },
		{ "Router", DHCP_OPTION_ROUTER, 3 },
		{ "TimeServer", DHCP_OPTION_TIMESERVER, 3 },
		{ "NameServer", DHCP_OPTION_NAMESERVER, 3 },
		{ "DomainServer", DHCP_OPTION_DNS, 3 },
		{ "LogServer", DHCP_OPTION_LOGSERVER, 3 },
		{ "QuotesServer", DHCP_OPTION_COOKIESERVER, 3 },
		{ "LPRServer", DHCP_OPTION_LPRSERVER, 3 },
		{ "ImpressServer", DHCP_OPTION_IMPRESSSERVER, 3 },
		{ "RLPServer", DHCP_OPTION_RESLOCSERVER, 3 },
		{ "Hostname", DHCP_OPTION_HOSTNAME, 1 },
		{ "BootFileSize", DHCP_OPTION_BOOTFILESIZE, 5 },
		{ "MeritDumpFile", DHCP_OPTION_MERITDUMP, 1 },
		{ "DomainName", DHCP_OPTION_DOMAINNAME, 1 },
		{ "SwapServer", DHCP_OPTION_SWAPSERVER, 3 },
		{ "RootPath", DHCP_OPTION_ROOTPATH, 1 },
		{ "ExtensionFile", DHCP_OPTION_EXTSPATH, 1 },
		{ "ForwardOn/Off", DHCP_OPTION_IPFORWARD, 7 },
		{ "SrcRteOn/Off", DHCP_OPTION_NONLOCALSR, 7 },
		{ "PolicyFilter", DHCP_OPTION_POLICYFILTER, 8 },
		{ "MaxDGAssembly", DHCP_OPTION_MAXREASSEMBLE, 2 },
		{ "DefaultIPTTL", DHCP_OPTION_IPTTL, 6 },
		{ "MTUTimeout", DHCP_OPTION_PATHMTUAGING, 4 },
		{ "MTUPlateau", DHCP_OPTION_PATHMTUPLATEAU, 2 },
		{ "MTUInterface", DHCP_OPTION_INTERFACEMTU, 5 },
		{ "MTUSubnet", DHCP_OPTION_SUBNETSLOCAL, 7 },
		{ "BroadcastAddress", DHCP_OPTION_BCASTADDRESS, 3 },
		{ "MaskDiscovery", DHCP_OPTION_MASKDISCOVERY, 7 },
		{ "MaskSupplier", DHCP_OPTION_MASKSUPPLIER, 7 },
		{ "RouterDiscovery", DHCP_OPTION_ROUTERDISCOVERY, 7 },
		{ "RouterRequest", DHCP_OPTION_ROUTERSOLIC, 3 },
		{ "StaticRoute", DHCP_OPTION_STATICROUTE, 8 },
		{ "Trailers", DHCP_OPTION_TRAILERENCAPS, 7 },
		{ "ARPTimeout", DHCP_OPTION_ARPTIMEOUT, 4 },
		{ "Ethernet", DHCP_OPTION_ETHERNETENCAPS, 7 },
		{ "DefaultTCPTTL", DHCP_OPTION_TCPTTL, 6 },
		{ "KeepaliveTime", DHCP_OPTION_TCPKEEPALIVEINT, 4 },
		{ "KeepaliveData", DHCP_OPTION_TCPKEEPALIVEGRBG, 7 },
		{ "NISDomain", DHCP_OPTION_NISDOMAIN, 1 },
		{ "NISServers", DHCP_OPTION_NISSERVERS, 3 },
		{ "NTPServers", DHCP_OPTION_NTPSERVERS, 3 },
		{ "NETBIOSNameSrv", DHCP_OPTION_NETBIOSNAMESERV, 3 },
		{ "NETBIOSDistSrv", DHCP_OPTION_NETBIOSDGDIST, 3 },
		{ "NETBIOSNodeType", DHCP_OPTION_NETBIOSNODETYPE, 6 },
		{ "NETBIOSScope", DHCP_OPTION_NETBIOSSCOPE, 1 },
		{ "XWindowFont", DHCP_OPTION_X11FONTS, 3 },
		{ "XWindowManager", DHCP_OPTION_X11DISPLAYMNGR, 3 },
		{ "AddressTime", DHCP_OPTION_IPADDRLEASE, 4 },
		{ "RenewalTime", DHCP_OPTION_RENEWALTIME, 4 },
		{ "RebindingTime", DHCP_OPTION_REBINDINGTIME, 4 },
		{ "NetWareIPDomain", DHCP_OPTION_NETWARE_IPDOMAIN, 1 },
		{ "NetWareIPOption", DHCP_OPTION_NETWARE_IPOPTION, 2 },
		{ "NISDomainName", DHCP_OPTION_NISPLUSDOMAIN, 1 },
		{ "NISServerAddr", DHCP_OPTION_NISPLUSSERVERS, 3 },
		{ "TFTPServerName", DHCP_OPTION_TFTPSERVER, 1 },
		{ "BootFileName", DHCP_OPTION_BP_FILE, 1 },
		{ "BootFileOption", DHCP_OPTION_BOOTFILE, 1 },
		{ "HomeAgentAddrs", DHCP_OPTION_MOBILEIPHOME, 3 },
		{ "SMTPServer", DHCP_OPTION_SMTPSERVER, 3 },
		{ "POP3Server", DHCP_OPTION_POP3SERVER, 3 },
		{ "NNTPServer", DHCP_OPTION_NNTPSERVER, 3 },
		{ "WWWServer", DHCP_OPTION_WWWSERVER, 3 },
		{ "FingerServer", DHCP_OPTION_FINGERSERVER, 3 },
		{ "IRCServer", DHCP_OPTION_IRCSERVER, 3 },
		{ "StreetTalkServer", DHCP_OPTION_STSERVER, 3 },
		{ "STDAServer", DHCP_OPTION_STDASERVER, 3 },
		{ "DirectoryAgent", DHCP_OPTION_SLPDIRAGENT, 1 },
		{ "ServiceScope", DHCP_OPTION_SLPDIRSCOPE, 1 },
		{ "iSNS", DHCP_OPTION_I_SNS, 1 },
		{ "NDSServers", DHCP_OPTION_NDSSERVERS, 3 },
		{ "NDSTreeName", DHCP_OPTION_NDSTREENAME, 1 },
		{ "NDSContext", DHCP_OPTION_NDSCONTEXT, 1 },
		{ "LDAP", DHCP_OPTION_LDAP, 1 },
		{ "PCode", DHCP_OPTION_P_CODE, 1 },
		{ "TCode", DHCP_OPTION_T_CODE, 1 },
		{ "NetInfoAddress", DHCP_OPTION_NETINFOADDRESS, 3 },
		{ "NetInfoTag", DHCP_OPTION_NETINFOTAG, 1 },
		{ "URL", DHCP_OPTION_URL, 1 },
		{ "AutoConfig", DHCP_OPTION_AUTO_CONFIG, 7 },
		{ "NameServiceSearch", DHCP_OPTION_NAMESERVICESEARCH, 2 },
		{ "SubnetSelectionOption", DHCP_OPTION_SUBNETSELECTION, 3 },
		{ "DomainSearch", DHCP_OPTION_DOMAINSEARCH, 1 },
		{ "SIPServersDHCPOption", DHCP_OPTION_SIPSERVERSDHCP, 1 },
		{ "ClasslessStaticRouteOption", DHCP_OPTION_CLASSLESSSTATICROUTE, 1 },
		{ "CCC", DHCP_OPTION_CCC, 1 },
		{ "TFTPServerIPaddress", DHCP_OPTION_TFPTSERVERIPADDRESS, 3 },
		{ "CallServerIPaddress", DHCP_OPTION_CALLSERVERIPADDRESS, 3 },
		{ "DiscriminationString", DHCP_OPTION_DISCRIMINATIONSTRING, 1 },
		{ "RemoteStatisticsServerIPAddress", DHCP_OPTION_REMOTESTATISTICSSERVER, 3 },
		{ "HTTPProxyPhone", DHCP_OPTION_HTTPPROXYFORPHONE_SPEC, 3 },
		{ "OPTION_CAPWAP_AC_V4", 138, 1 },
		{ "OPTIONIPv4_AddressMoS", 139, 1 },
		{ "OPTIONIPv4_FQDNMoS", 140, 1 },
		{ "SIPUAServiceDomains", 141, 1 },
		{ "OPTIONIPv4_AddressANDSF", 142, 1 },
		{ "IPTelephone", 176, 1 },
		{ "ConfigurationFile", 209, 1 },
		{ "PathPrefix", 210, 1 },
		{ "OPTION_6RD", 212, 1 },
		{ "OPTION_V4_ACCESS_DOMAIN", 213, 1 },
        { "NextServer", DHCP_OPTION_NEXTSERVER, 3},
        { "BootTFTPServer", DHCP_OPTION_NEXTSERVER, 3}
    };

int main(int argc, char **argv)
{
    signal(SIGINT, catch_int);
    signal(SIGABRT, catch_int);
    signal(SIGTERM, catch_int);
    signal(SIGQUIT, catch_int);
    signal(SIGTSTP, catch_int);
    signal(SIGHUP, catch_int);

    //printf("%i\n", argc);
/*
	logBuff[0] = 0;
	char *ds = strrchr(argv[0], '/');

	if (ds)
		ds++;
	else
		ds = argv[0];

	sprintf(tempbuff, "ps -e | grep -v grep | grep -v \"rc.%s\" | grep \"%s$\" | head -1 | awk '{ print $1 }'", ds, ds);
	FILE *p = popen(tempbuff,"r");

	if (p)
	{
		while (fgets(tempbuff, sizeof(tempbuff), p))
		{
			if (atoi(tempbuff) != getpid())
			{
				sprintf(logBuff, "Error: %s is already running, Pid = %s", ds, tempbuff);
				break;
			}
		}
		pclose(p);
	}
*/

    for (int i = 1; i < argc; i++)
    {
        if (!strcasecmp(argv[i], "-v"))
            verbatim = true;
        else if (!strcmp(argv[i], "-i") && argc > i + 1 && argv[i + 1][0] != '-' )
        {
            myTrim(iniFile, argv[i + 1]);
            i++;
        }
        else if (!strcmp(argv[i], "-l") && argc > i + 1 && argv[i + 1][0] != '-' )
        {
            myTrim(logFile, argv[i + 1]);
            i++;
        }
        else if (!strcmp(argv[i], "-s") && argc > i + 1 && argv[i + 1][0] != '-' )
        {
            myTrim(leaFile, argv[i + 1]);
            i++;
        }
        else if (!strncasecmp(argv[i], "-i", 2))
            myTrim(iniFile, argv[i] + 2);
        else if (!strncasecmp(argv[i], "-l", 2))
            myTrim(logFile, argv[i] + 2);
        else if (!strncasecmp(argv[i], "-s", 2))
            myTrim(leaFile, argv[i] + 2);
        else
            sprintf(logBuff, "Error: Invalid Argument %s", argv[i]);
    }

    if (!leaFile[0])
        strcpy(leaFile, "/tmp/opendhcp.state");

    if (!iniFile[0])
        strcpy(iniFile, "/etc/opendhcp.ini");

    if (verbatim)
    {
		if (logBuff[0])
		{
			printf("%s\n", logBuff);
			exit(EXIT_FAILURE);
		}

        if (getuid())
        {
        	printf("Error: Only root should run this program\n");
			exit(EXIT_FAILURE);
        }

        //printf("%i\n",sizeof(data7));
	    memset(&cfig, 0, sizeof(cfig));
        cfig.ppid = getpid();
		pthread_t threadId;
		pthread_attr_t attr;
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		int errcode = pthread_create(&threadId, &attr, init, NULL);
		pthread_attr_destroy(&attr);

		if(errcode)
		{
			if (cfig.dhcpLogLevel >= 1)
			{
				sprintf(logBuff, "Init thread Creation Failed");
				logDHCPMess(logBuff, 1);
			}
			exit(EXIT_FAILURE);
		}

		while (kRunning)
		{
			network.busy = false;

			if (kRunning && (!network.dhcpConn[0].ready || !network.ready))
			{
				sleep(1);
				continue;
			}

			FD_ZERO(&readfds);
			tv.tv_sec = 20;
			tv.tv_usec = 0;

			if (network.httpConn.ready)
				FD_SET(network.httpConn.sock, &readfds);

			for (int i = 0; i < MAX_SERVERS && network.dhcpConn[i].ready; i++)
				FD_SET(network.dhcpConn[i].sock, &readfds);

			FD_SET(network.dhcpListener.sock, &readfds);

			if (cfig.replication)
				FD_SET(cfig.dhcpReplConn.sock, &readfds);

			//printf("%i\n",select(network.maxFD, &readfds, NULL, NULL, &tv));
			if (select(network.maxFD, &readfds, NULL, NULL, &tv))
			{
				t = time(NULL);

				if (network.ready)
				{
					network.busy = true;

					if (network.httpConn.ready && FD_ISSET(network.httpConn.sock, &readfds))
					{
						data19 *req = (data19*)calloc(1, sizeof(data19));

						if (req)
						{
							req->sockLen = sizeof(req->remote);
							errno = 0;
							req->sock = accept(network.httpConn.sock, (sockaddr*)&req->remote, &req->sockLen);

							if (req->sock == INVALID_SOCKET)
							{
								sprintf(logBuff, "Accept Failed, Error=%s\n", strerror(errno));
								logDHCPMess(logBuff, 1);
								free(req);
							}
							else
								procHTTP(req);
						}
						else
						{
							sprintf(logBuff, "Memory Error");
							logDHCPMess(logBuff, 0);
						}
					}

					for (int i = 0; i < MAX_SERVERS && network.dhcpConn[i].ready; i++)
					{
						if (FD_ISSET(network.dhcpConn[i].sock, &readfds) && gdmess(&dhcpr, i) && sdmess(&dhcpr))
							alad(&dhcpr);
					}

					if (FD_ISSET(network.dhcpListener.sock, &readfds) && gdmess(&dhcpr, 255) && sdmess(&dhcpr))
						alad(&dhcpr);

					if (cfig.replication && FD_ISSET(cfig.dhcpReplConn.sock, &readfds))
					{
						//printf("Repl\n");
						dhcpr.sockLen = sizeof(dhcpr.remote);
						dhcpr.bytes = recvfrom(cfig.dhcpReplConn.sock,
											   dhcpr.raw,
											   sizeof(dhcpr.raw),
											   0,
											   (sockaddr*)&dhcpr.remote,
											   &dhcpr.sockLen);
					}
				}
			}
			else
				t = time(NULL);
		}
    }
    else
    {
		if(logBuff[0])
		{
			syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), logBuff);
			exit(EXIT_FAILURE);
		}

        if(getuid())
        {
            syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), "Only root should run this program");
			exit(EXIT_FAILURE);
        }

        /* Our process ID and Session ID */
        pid_t sid, pid;

        /* Fork off the parent process */
        pid = fork();
        if (pid < 0)
        {
            exit(EXIT_FAILURE);
        }
        /* If we got a good PID, then
        we can exit the parent process. */
        if (pid > 0)
        {
            exit(EXIT_SUCCESS);
        }

        /* Change the file mode mask */
        umask(0);

        /* Open any logs here */

        /* Create a new SID for the child process */
        sid = setsid();
        if (sid < 0)
        {
            /* Log the failure */
            exit(EXIT_FAILURE);
        }

        /* Change the current working directory */
        if ((chdir("/")) < 0)
        {
            /* Log the failure */
            exit(EXIT_FAILURE);
        }

        /* Close out the standard file descriptors */
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);

        /* Daemon-specific initialization goes here */
	    memset(&cfig, 0, sizeof(cfig));
        cfig.ppid = getpid();
        data9 dhcpr;
		pthread_t threadId;
		pthread_attr_t attr;
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		int errcode = pthread_create(&threadId, &attr, init, NULL);
		pthread_attr_destroy(&attr);

		if(errcode)
		{
			if (cfig.dhcpLogLevel >= 1)
			{
				sprintf(logBuff, "init thread Creation Failed");
				logDHCPMess(logBuff, 1);
			}
			exit(EXIT_FAILURE);
		}

		/* The Big Loop */
		while (kRunning)
		{
			network.busy = false;

			if (kRunning && (!network.dhcpConn[0].ready || !network.ready))
			{
				sleep(1);
				continue;
			}

			FD_ZERO(&readfds);
			tv.tv_sec = 20;
			tv.tv_usec = 0;

			if (network.httpConn.ready)
				FD_SET(network.httpConn.sock, &readfds);

			for (int i = 0; i < MAX_SERVERS && network.dhcpConn[i].ready; i++)
				FD_SET(network.dhcpConn[i].sock, &readfds);

			FD_SET(network.dhcpListener.sock, &readfds);

			if (cfig.replication)
				FD_SET(cfig.dhcpReplConn.sock, &readfds);

			//printf("%i\n",select(network.maxFD, &readfds, NULL, NULL, &tv));
			if (select(network.maxFD, &readfds, NULL, NULL, &tv))
			{
				t = time(NULL);

				if (network.ready)
				{
					network.busy = true;

					if (network.httpConn.ready && FD_ISSET(network.httpConn.sock, &readfds))
					{
						data19 *req = (data19*)calloc(1, sizeof(data19));

						if (req)
						{
							req->sockLen = sizeof(req->remote);
							errno = 0;
							req->sock = accept(network.httpConn.sock, (sockaddr*)&req->remote, &req->sockLen);

							if (req->sock == INVALID_SOCKET)
							{
								sprintf(logBuff, "Accept Failed, Error=%s\n", strerror(errno));
								logDHCPMess(logBuff, 1);
								free(req);
							}
							else
								procHTTP(req);
						}
						else
						{
							sprintf(logBuff, "Memory Error");
							logDHCPMess(logBuff, 0);
						}
					}

					for (int i = 0; i < MAX_SERVERS && network.dhcpConn[i].ready; i++)
					{
						if (FD_ISSET(network.dhcpConn[i].sock, &readfds) && gdmess(&dhcpr, i) && sdmess(&dhcpr))
							alad(&dhcpr);
					}

					if (FD_ISSET(network.dhcpListener.sock, &readfds) && gdmess(&dhcpr, 255) && sdmess(&dhcpr))
						alad(&dhcpr);

					if (cfig.replication && FD_ISSET(cfig.dhcpReplConn.sock, &readfds))
					{
						//printf("Repl\n");
						dhcpr.sockLen = sizeof(dhcpr.remote);

						dhcpr.bytes = recvfrom(cfig.dhcpReplConn.sock,
											   dhcpr.raw,
											   sizeof(dhcpr.raw),
											   0,
											   (sockaddr*)&dhcpr.remote,
											   &dhcpr.sockLen);
					}
				}
			}
			else
				t = time(NULL);
		}
    }

    closeConn();

	if (cfig.dhcpReplConn.ready)
		close(cfig.dhcpReplConn.sock);
}

void closeConn()
{
	if (network.dhcpListener.ready)
		close(network.dhcpListener.sock);

	if (network.httpConn.ready)
		close(network.httpConn.sock);

	for (int i = 0; i < MAX_SERVERS && network.dhcpConn[i].loaded; i++)
	{
		if (network.dhcpConn[i].ready)
			close(network.dhcpConn[i].sock);
	}
}

void catch_int(int sig_num)
{
	//printf("%u=%u\n", cfig.ppid, getpid());
    if (cfig.ppid == getpid())
    {
		network.ready = false;
		kRunning = false;
		sprintf(logBuff, "Closing Network Connections...");
		logDHCPMess(logBuff, 1);

		closeConn();

		if (cfig.dhcpReplConn.ready)
			close(cfig.dhcpReplConn.sock);

		sprintf(logBuff, "Open DHCP Server Stopped !\n");
		logDHCPMess(logBuff, 1);

		if (cfig.logfile)
			fclose(cfig.logfile);

		close(cfig.fixedSocket);

		exit(EXIT_SUCCESS);
	}
}

MYWORD fUShort(void *raw)
{
	return ntohs(*((MYWORD*)raw));
}

MYDWORD fULong(void *raw)
{
	return ntohl(*((MYDWORD*)raw));
}

MYDWORD fIP(void *raw)
{
	return (*((MYDWORD*)raw));
}

MYBYTE pUShort(void *raw, MYWORD data)
{
	*((MYWORD*)raw) = htons(data);
	return sizeof(MYWORD);
}

MYBYTE pULong(void *raw, MYDWORD data)
{
	*((MYDWORD*)raw) = htonl(data);
	return sizeof(MYDWORD);
}

MYBYTE pIP(void *raw, MYDWORD data)
{
	*((MYDWORD*)raw) = data;
	return sizeof(MYDWORD);
}

void procHTTP(data19 *req)
{
	req->ling.l_onoff = 1; //0 = off (l_linger ignored), nonzero = on
	req->ling.l_linger = 30; //0 = discard data, nonzero = wait for data sent
	setsockopt(req->sock, SOL_SOCKET, SO_LINGER, (const char*)&req->ling, sizeof(req->ling));

	dhcpMap::iterator p;
	int ind = 0;
	MYDWORD iip = 0;
	data7 *dhcpEntry = NULL;
	//data7 *cache = NULL;
	MYDWORD memSize = 2048 + (135 * dhcpCache.size()) + ((cfig.dhcpSize - dhcpCache.size()) * 26);
	req->buffer = (char*)calloc(1, memSize);

	if (!req->buffer)
	{
		sprintf(logBuff, "Memory Error");
		logDHCPMess(logBuff, 1);
		closesocket(req->sock);
		free(req);
		return;
	}

	errno = 0;
	req->bytes = recv(req->sock, req->buffer, memSize, 0);
	//errno = WSAGetLastError();

	if (errno)
	{
		sprintf(logBuff, "Client %s, HTTP Message Receive failed, WSAError %u", IP2String(tempbuff, req->remote.sin_addr.s_addr), errno);
		logDHCPMess(logBuff, 1);
		free(req->buffer);
		closesocket(req->sock);
		free(req);
		return;
	}
	else if (cfig.dhcpLogLevel >= 2)
	{
		sprintf(logBuff, "Client %s, HTTP Request Received", IP2String(tempbuff, req->remote.sin_addr.s_addr));
		logDHCPMess(logBuff, 2);
		//printf("%s\n", req->buffer);
	}

	char *fp = req->buffer;
	char *maxData = req->buffer + (memSize - 512);
	tm *ttm = gmtime(&t);
	strftime(tempbuff, sizeof(tempbuff), "%a, %d %b %Y %H:%M:%S GMT", ttm);
	fp += sprintf(fp, send200, tempbuff, tempbuff);
	char *contentStart = fp;
	fp += sprintf(fp, "<html>\n<head>\n<meta http-equiv=\"refresh\" content=\"60\">\n<meta http-equiv=\"cache-control\" content=\"no-cache\">\n</head>\n");
	fp += sprintf(fp, "<body bgcolor=\"#cccccc\">\n<h2>%s</h2>\n", sVersion);
	//fp += sprintf(fp, "<body bgcolor=\"#cccccc\">\n<h2>Host: %s</h2>\n", cfig.servername);
	fp += sprintf(fp, "<table border=\"1\" width=\"640\" bgcolor=\"#b8b8b8\">\n");

	if (cfig.replication)
	{
		fp += sprintf(fp, "<tr><th colspan=\"5\"><font size=\"5\"><i>Active Leases</i></font></th></tr>\n");
		fp += sprintf(fp, "<tr><th>Mac Address</th><th>IP</th><th>Lease Expiry</th><th>Hostname (first 20 chars)</th><th>Server</th></tr>\n");
	}
	else
	{
		fp += sprintf(fp, "<tr><th colspan=\"4\"><font size=\"5\"><i>Active Leases</i></font></th></tr>\n");
		fp += sprintf(fp, "<tr><th>Mac Address</th><th>IP</th><th>Lease Expiry</th><th>Hostname (first 20 chars)</th></tr>\n");
	}

	for (p = dhcpCache.begin(); p != dhcpCache.end() && fp < maxData; p++)
	{
		if ((dhcpEntry = p->second) && dhcpEntry->display && dhcpEntry->expiry >= t)
		{
			fp += sprintf(fp, "<tr>");
			fp += sprintf(fp, line200, dhcpEntry->mapname);
			fp += sprintf(fp, line200, IP2String(tempbuff, dhcpEntry->ip));

			if (dhcpEntry->expiry >= INT_MAX)
				fp += sprintf(fp, line200, "Infinity");
			else
			{
				tm *ttm = localtime(&dhcpEntry->expiry);
				strftime(tempbuff, sizeof(tempbuff), "%d-%b-%y %X", ttm);
				fp += sprintf(fp, line200, tempbuff);
			}

			if (dhcpEntry->hostname)
			{
				if (strlen(dhcpEntry->hostname) <= 20)
					fp += sprintf(fp, line200, dhcpEntry->hostname);
				else
				{
					memcpy(tempbuff, dhcpEntry->hostname, 20);
					tempbuff[20] = 0;
					fp += sprintf(fp, line200, tempbuff);
				}
			}
			else
				fp += sprintf(fp, line200, "&nbsp;");

			if (cfig.replication)
			{
				if (dhcpEntry->hanged && cfig.replication == 1)
					fp += sprintf(fp, line200, "Primary");
				else if (dhcpEntry->hanged && cfig.replication == 2)
					fp += sprintf(fp, line200, "Secondary");
				else if (cfig.replication == 1)
					fp += sprintf(fp, line200, "Secondary");
				else
					fp += sprintf(fp, line200, "Primary");
			}

			fp += sprintf(fp, "</tr>\n");
		}
	}

	fp += sprintf(fp, "</table>\n<br>\n<table border=\"1\" width=\"640\" bgcolor=\"#b8b8b8\">\n");
	fp += sprintf(fp, "<tr><th colspan=\"5\"><font size=\"5\"><i>Free Dynamic Leases</i></font></th></tr>\n");
	MYBYTE colNum = 0;

	for (char rangeInd = 0; rangeInd < MAX_RANGES && cfig.dhcpRanges[rangeInd].rangeStart && fp < maxData; rangeInd++)
	{
		for (ind = 0, iip = cfig.dhcpRanges[rangeInd].rangeStart; iip <= cfig.dhcpRanges[rangeInd].rangeEnd; iip++, ind++)
		{
			if (cfig.dhcpRanges[rangeInd].expiry[ind] < t)
			{
				if (!colNum)
				{
					fp += sprintf(fp, "<tr>");
					colNum = 1;
				}
				else if (colNum < 5)
					colNum++;
				else
				{
					fp += sprintf(fp, "</tr>\n<tr>");
					colNum = 1;
				}

				fp += sprintf(fp, line200, IP2String(tempbuff, htonl(iip)));
			}
		}
	}

	if (colNum)
		fp += sprintf(fp, "</tr>\n");

	fp += sprintf(fp, "</table>\n<br>\n<table border=\"1\" width=\"640\" bgcolor=\"#b8b8b8\">\n");
	fp += sprintf(fp, "<tr><th colspan=\"4\"><font size=\"5\"><i>Free Static Leases</i></font></th></tr>\n");
	fp += sprintf(fp, "<tr><th>Mac Address</th><th>IP</th><th>Mac Address</th><th>IP</th></tr>\n");
	colNum = 0;

	for (p = dhcpCache.begin(); p != dhcpCache.end() && fp < maxData; p++)
	{
		if ((dhcpEntry = p->second) && dhcpEntry->fixed && dhcpEntry->expiry < t)
		{
			if (!colNum)
			{
				fp += sprintf(fp, "<tr>");
				colNum = 1;
			}
			else if (colNum == 1)
			{
				colNum = 2;
			}
			else if (colNum == 2)
			{
				fp += sprintf(fp, "</tr>\n<tr>");
				colNum = 1;
			}

			fp += sprintf(fp, line200, dhcpEntry->mapname);
			fp += sprintf(fp, line200, IP2String(tempbuff, dhcpEntry->ip));
		}
	}

	if (colNum)
		fp += sprintf(fp, "</tr>\n");

	fp += sprintf(fp, "</table>\n</body>\n</html>");
	memcpy((contentStart - 12), tempbuff, sprintf(tempbuff, "%u", (fp - contentStart)));
	req->bytes = fp - req->buffer;

	pthread_t threadId;
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	int errcode = pthread_create(&threadId, &attr, sendHTTP, (void*)req);
	pthread_attr_destroy(&attr);

	if(errcode)
	{
		if (cfig.dhcpLogLevel >= 1)
		{
			sprintf(logBuff, "Thread Creation Failed");
			logDHCPMess(logBuff, 1);
		}
		send(req->sock, req->buffer, req->bytes, 0);
		closesocket(req->sock);
		free(req->buffer);
		free(req);
	}

	return;
}

void *sendHTTP(void *lpParam)
{
	//printf("Here\n");
	data19 *req = (data19*)lpParam;
	send(req->sock, req->buffer, req->bytes, 0);
	sleep(20);
	closesocket(req->sock);
	free(req->buffer);
	free(req);
	pthread_exit(NULL);
}


bool checkRange(char rangeInd, bool macFound, bool vendFound, bool userFound)
{
	if (!cfig.hasFilter)
		return true;

	MYBYTE rangeSetInd = cfig.dhcpRanges[rangeInd].rangeSetInd;
	data14 *rangeSet = &cfig.rangeSet[rangeSetInd];
	//printf("checkRange entering, rangeInd=%i rangeSetInd=%i\n", rangeInd, rangeSetInd);
	//printf("checkRange entered, macFound=%i vendFound=%i userFound=%i\n", macFound, vendFound, userFound);

	if((!macFound && !rangeSet->macSize[0]) || (macFound && cfig.macArray[rangeSetInd]))
		if((!vendFound && !rangeSet->vendClassSize[0]) || (vendFound && cfig.vendArray[rangeSetInd]))
			if((!userFound && !rangeSet->userClassSize[0]) || (userFound && cfig.userArray[rangeSetInd]))
				return true;

	//printf("checkRange, returning false rangeInd=%i rangeSetInd=%i\n", rangeInd, rangeSetInd);
	return false;
}

MYDWORD resad(data9 *req)
{
	//printf("In resad\n");

	MYDWORD minRange = 0;
	MYDWORD maxRange = 0;

	if (req->dhcpp.header.bp_giaddr)
	{
		lockIP(req->dhcpp.header.bp_giaddr);
		lockIP(req->remote.sin_addr.s_addr);
	}

	if (chad(req) && req->dhcpEntry->fixed)
	{
		//req->dhcpEntry->no_route = 1;
		setLeaseExpiry(req->dhcpEntry, 20);
		return req->dhcpEntry->ip;
	}


	MYDWORD iipNew = 0;
	MYDWORD iipExp = 0;
	MYDWORD rangeStart = 0;
	MYDWORD rangeEnd = 0;
	char rangeInd = -1;
	bool rangeFound = false;
	bool macFound = false;
	bool vendFound = false;
	bool userFound = false;
	memset(cfig.macArray, 0, sizeof(cfig.macArray));
	memset(cfig.vendArray, 0, sizeof(cfig.vendArray));
	memset(cfig.userArray, 0, sizeof(cfig.userArray));


	if (cfig.hasFilter)
	{
		for (MYBYTE rangeSetInd = 0; rangeSetInd < MAX_RANGES && cfig.rangeSet[rangeSetInd].active; rangeSetInd++)
		{
			data14 *rangeSet = &cfig.rangeSet[rangeSetInd];

			for (MYBYTE i = 0; i < 32 && rangeSet->macSize[i]; i++)
			{
				if(memcmp(req->dhcpp.header.bp_chaddr, rangeSet->macStart[i], rangeSet->macSize[i]) >= 0 && memcmp(req->dhcpp.header.bp_chaddr, rangeSet->macEnd[i], rangeSet->macSize[i]) <= 0)
				{
					cfig.macArray[rangeSetInd] = true;
					macFound = true;
					//printf("mac Found, rangeSetInd=%i\n", rangeSetInd);
					break;
				}
			}

			for (MYBYTE i = 0; i < 32 && req->vendClassSize && rangeSet->vendClassSize[i]; i++)
			{
				if(rangeSet->vendClassSize[i] == req->vendClassSize && !memcmp(req->vendClass, rangeSet->vendClass[i], rangeSet->vendClassSize[i]))
				{
					cfig.vendArray[rangeSetInd] = true;
					vendFound = true;
					//printf("vend Found, rangeSetInd=%i\n", rangeSetInd);
					break;
				}
			}

			for (MYBYTE i = 0; i < 32 && req->userClassSize && rangeSet->userClassSize[i]; i++)
			{
				if(rangeSet->userClassSize[i] == req->userClassSize && !memcmp(req->userClass, rangeSet->userClass[i], rangeSet->userClassSize[i]))
				{
					cfig.userArray[rangeSetInd] = true;
					userFound = true;
					//printf("user Found, rangeSetInd=%i\n", rangeSetInd);
					break;
				}
			}
		}

	}

//	printArray("macArray", (char*)cfig.macArray);
//	printArray("vendArray", (char*)cfig.vendArray);
//	printArray("userArray", (char*)cfig.userArray);

	if (req->dhcpEntry || chad(req))
	{
		int ind = getIndex(req->dhcpEntry->rangeInd, req->dhcpEntry->ip);

		if (cfig.dhcpRanges[req->dhcpEntry->rangeInd].dhcpEntry[ind] == req->dhcpEntry)
		{
			if (checkRange(req->dhcpEntry->rangeInd, macFound, vendFound, userFound))
			{
				MYDWORD mask = INADDR_NONE;

				if (req->dhcpEntry->rangeInd >= 0)
					mask = cfig.dhcpRanges[req->dhcpEntry->rangeInd].mask;
				else
					mask = cfig.mask;

				if (req->dhcpp.header.bp_giaddr)
					calcRangeLimits(req->dhcpp.header.bp_giaddr, mask, &minRange, &maxRange);
//				else if (htonl(network.dhcpConn[req->sockInd].mask) > htonl(mask))
//					calcRangeLimits(network.dhcpConn[req->sockInd].server, network.dhcpConn[req->sockInd].mask, &minRange, &maxRange);
				else
					calcRangeLimits(network.dhcpConn[req->sockInd].server, mask, &minRange, &maxRange);

				if (htonl(req->dhcpEntry->ip) >= minRange && htonl(req->dhcpEntry->ip) <= maxRange)
				{
					setLeaseExpiry(req->dhcpEntry, 20);
					return req->dhcpEntry->ip;
				}
			}
		}
	}

	if (!iipNew && req->reqIP)
	{
		char k = getRangeInd(req->reqIP);

		if (k >= 0)
		{
			if (checkRange(k, macFound, vendFound, userFound))
			{
				data13 *range = &cfig.dhcpRanges[k];
				int ind = getIndex(k, req->reqIP);

				if (ind >= 0 && range->expiry[ind] <= t)
				{
					if (req->dhcpp.header.bp_giaddr)
						calcRangeLimits(req->dhcpp.header.bp_giaddr, range->mask, &minRange, &maxRange);
//					else if (htonl(network.dhcpConn[req->sockInd].mask) > htonl(range->mask))
//						calcRangeLimits(network.dhcpConn[req->sockInd].server, network.dhcpConn[req->sockInd].mask, &minRange, &maxRange);
					else
						calcRangeLimits(network.dhcpConn[req->sockInd].server, range->mask, &minRange, &maxRange);

					MYDWORD iip = htonl(req->reqIP);

					if (iip >= minRange && iip <= maxRange)
					{
						iipNew = iip;
						rangeInd = k;
					}
				}
			}
		}
	}


	for (char k = 0; !iipNew && k < MAX_RANGES && cfig.dhcpRanges[k].rangeStart; k++)
	{
		if (checkRange(k, macFound, vendFound, userFound))
		{
			data13 *range = &cfig.dhcpRanges[k];
			rangeStart = range->rangeStart;
			rangeEnd = range->rangeEnd;

			if (req->dhcpp.header.bp_giaddr)
				calcRangeLimits(req->dhcpp.header.bp_giaddr, range->mask, &minRange, &maxRange);
//			else if (htonl(network.dhcpConn[req->sockInd].mask) > htonl(range->mask))
//				calcRangeLimits(network.dhcpConn[req->sockInd].server, network.dhcpConn[req->sockInd].mask, &minRange, &maxRange);
			else
				calcRangeLimits(network.dhcpConn[req->sockInd].server, range->mask, &minRange, &maxRange);

			if (rangeStart < minRange)
				rangeStart = minRange;

			if (rangeEnd > maxRange)
				rangeEnd = maxRange;

			if (rangeStart <= rangeEnd)
			{
				rangeFound = true;

				if (cfig.replication == 2)
				{
					for (MYDWORD m = rangeEnd; m >= rangeStart; m--)
					{
						if (!range->expiry[m - range->rangeStart])
						{
							iipNew = m;
							rangeInd = k;
							break;
						}
						else if (!iipExp && range->expiry[m - range->rangeStart] < t)
						{
							iipExp = m;
							rangeInd = k;
						}
					}
				}
				else
				{
					for (MYDWORD m = rangeStart; m <= rangeEnd; m++)
					{
						if (!range->expiry[m - range->rangeStart])
						{
							iipNew = m;
							rangeInd = k;
							break;
						}
						else if (!iipExp && range->expiry[m - range->rangeStart] < t)
						{
							iipExp = m;
							rangeInd = k;
						}
					}
				}
			}
		}
	}


	if (!iipNew && iipExp)
			iipNew = iipExp;

	if (iipNew)
	{
		if (!req->dhcpEntry)
		{
			req->dhcpEntry = (data7*)calloc(1, sizeof(data7));

			if (!req->dhcpEntry)
			{
				sprintf(logBuff, "Memory Allocation Error");
				logDHCPMess(logBuff, 1);
				return 0;
			}

			req->dhcpEntry->mapname = cloneString(req->chaddr);

			if (!req->dhcpEntry->mapname)
			{
				sprintf(logBuff, "Memory Allocation Error");
				logDHCPMess(logBuff, 1);
				return 0;
			}

			dhcpCache[req->dhcpEntry->mapname] = req->dhcpEntry;
		}

		req->dhcpEntry->ip = htonl(iipNew);
		req->dhcpEntry->rangeInd = rangeInd;
		setLeaseExpiry(req->dhcpEntry, 20);
		//req->dhcpEntry->source = network.dhcpConn[req->sockInd].server;

		return req->dhcpEntry->ip;
	}

	if (!iipNew)
	{
		if (cfig.dhcpLogLevel >= 1)
		{
			if (rangeFound)
			{
				if (req->dhcpp.header.bp_giaddr)
					sprintf(logBuff, "No free leases for DHCP discover for %s (%s) from RelayAgent %s", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_giaddr));
				else
					sprintf(logBuff, "No free leases for DHCP discover for %s (%s) from interface %s", req->chaddr, req->hostname, IP2String(tempbuff, network.dhcpConn[req->sockInd].server));
			}
			else
			{
				if (req->dhcpp.header.bp_giaddr)
					sprintf(logBuff, "No Matching DHCP Range for DHCP discover for %s (%s) from RelayAgent %s", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_giaddr));
				else
					sprintf(logBuff, "No Matching DHCP Range for DHCP discover for %s (%s) from interface %s", req->chaddr, req->hostname, IP2String(tempbuff, network.dhcpConn[req->sockInd].server));
			}
			logDHCPMess(logBuff, 1);
		}
	}
	return 0;
}

MYDWORD chad(data9 *req)
{
	req->dhcpEntry = findDHCPEntry(req->chaddr);
	//printf("dhcpEntry=%d\n", req->dhcpEntry);

	if (req->dhcpEntry && req->dhcpEntry->ip)
		return req->dhcpEntry->ip;
	else
		return 0;
}

MYDWORD sdmess(data9 *req)
{
	//printf("Request=%u Req IP=%s\n",req->req_type, IP2String(tempbuff, req->dhcpp.header.bp_ciaddr));
	//printf("In sdmess\n");

	if (req->req_type == DHCP_MESS_NONE)
	{
		req->dhcpp.header.bp_yiaddr = chad(req);

		if (!req->dhcpp.header.bp_yiaddr)
		{
			if (cfig.dhcpLogLevel >= 1)
			{
				sprintf(logBuff, "No Static Entry found for BOOTP request from Host %s", req->chaddr);
				logDHCPMess(logBuff, 1);
			}

			return 0;
		}
	}
	else if (req->req_type == DHCP_MESS_DECLINE)
	{
		if (req->dhcpp.header.bp_ciaddr && chad(req) == req->dhcpp.header.bp_ciaddr)
		{
			lockIP(req->dhcpp.header.bp_ciaddr);

			req->dhcpEntry->ip = 0;
			req->dhcpEntry->expiry = INT_MAX;

			if (cfig.dhcpLogLevel >= 1)
			{
				sprintf(logBuff, "IP Address %s declined by Host %s (%s), locked", IP2String(tempbuff, req->dhcpp.header.bp_ciaddr), req->chaddr, req->hostname);
				logDHCPMess(logBuff, 1);
			}
		}

		return 0;
	}
	else if (req->req_type == DHCP_MESS_RELEASE)
	{
		if (req->dhcpp.header.bp_ciaddr && chad(req) == req->dhcpp.header.bp_ciaddr)
		{
			req->lease = 0;
			updateDHCP(req);

			if (cfig.dhcpLogLevel >= 1)
			{
				sprintf(logBuff, "IP Address %s released by Host %s (%s)", IP2String(tempbuff, req->dhcpp.header.bp_ciaddr), req->chaddr, req->hostname);
				logDHCPMess(logBuff, 1);
			}
		}

		return 0;
	}
	else if (req->req_type == DHCP_MESS_INFORM)
	{
		//printf("repl0=%s\n", IP2String(tempbuff, cfig.zoneServers[0]));
		//printf("repl1=%s\n", IP2String(tempbuff, cfig.zoneServers[1]));
		//printf("IP=%s bytes=%u replication=%i\n", IP2String(tempbuff, req->remote.sin_addr.s_addr), req->bytes, cfig.replication);

		if ((cfig.replication == 1 && req->remote.sin_addr.s_addr == cfig.zoneServers[1]) || (cfig.replication == 2 && req->remote.sin_addr.s_addr == cfig.zoneServers[0]))
			recvRepl(req);

		return 0;
	}
	else if (req->req_type == DHCP_MESS_DISCOVER && strcasecmp(req->hostname, cfig.servername))
	{
		req->dhcpp.header.bp_yiaddr = resad(req);

		if (!req->dhcpp.header.bp_yiaddr)
			return 0;

		req->resp_type = DHCP_MESS_OFFER;
	}
	else if (req->req_type == DHCP_MESS_REQUEST)
	{
		//printf("%s\n", IP2String(tempbuff, req->dhcpp.header.bp_ciaddr));

		if (req->server)
		{
			if (req->server == network.dhcpConn[req->sockInd].server)
			{
				if (req->reqIP && req->reqIP == chad(req) && req->dhcpEntry->expiry > t)
				{
					req->resp_type = DHCP_MESS_ACK;
					req->dhcpp.header.bp_yiaddr = req->reqIP;
				}
				else if (req->dhcpp.header.bp_ciaddr && req->dhcpp.header.bp_ciaddr == chad(req) && req->dhcpEntry->expiry > t)
				{
					req->resp_type = DHCP_MESS_ACK;
					req->dhcpp.header.bp_yiaddr = req->dhcpp.header.bp_ciaddr;
				}
				else
				{
					req->resp_type = DHCP_MESS_NAK;
					req->dhcpp.header.bp_yiaddr = 0;

					if (cfig.dhcpLogLevel >= 1)
					{
						sprintf(logBuff, "DHCP Request from Host %s (%s) without Discover, NAKed", req->chaddr, req->hostname);
						logDHCPMess(logBuff, 1);
					}
				}
			}
			else
				return 0;
		}
		else if (req->dhcpp.header.bp_ciaddr && req->dhcpp.header.bp_ciaddr == chad(req) && req->dhcpEntry->expiry > t)
		{
			req->resp_type = DHCP_MESS_ACK;
			req->dhcpp.header.bp_yiaddr = req->dhcpp.header.bp_ciaddr;
		}
		else if (req->reqIP && req->reqIP == chad(req) && req->dhcpEntry->expiry > t)
		{
			req->resp_type = DHCP_MESS_ACK;
			req->dhcpp.header.bp_yiaddr = req->reqIP;
		}
		else
		{
			req->resp_type = DHCP_MESS_NAK;
			req->dhcpp.header.bp_yiaddr = 0;

			if (cfig.dhcpLogLevel >= 1)
			{
				sprintf(logBuff, "DHCP Request from Host %s (%s) without Discover, NAKed", req->chaddr, req->hostname);
				logDHCPMess(logBuff, 1);
			}
		}
	}
	else
		return 0;

	addOptions(req);
	int packSize = req->vp - (MYBYTE*)&req->dhcpp;
	packSize++;

	if (req->req_type == DHCP_MESS_NONE)
		packSize = req->messsize;

	if (req->dhcpp.header.bp_giaddr)
	{
		req->remote.sin_port = htons(IPPORT_DHCPS);
		req->remote.sin_addr.s_addr = req->dhcpp.header.bp_giaddr;
	}
	else if (req->dhcpp.header.bp_broadcast || !req->remote.sin_addr.s_addr || req->reqIP)
	{
		req->remote.sin_port = htons(IPPORT_DHCPC);
		req->remote.sin_addr.s_addr = INADDR_BROADCAST;
		req->dhcpp.header.bp_broadcast = 1;
	}
	else
	{
		req->remote.sin_port = htons(IPPORT_DHCPC);
		req->dhcpp.header.bp_broadcast = 0;
	}

	req->dhcpp.header.bp_op = BOOTP_REPLY;
	errno = 0;

	if (req->req_type == DHCP_MESS_DISCOVER && !req->dhcpp.header.bp_giaddr)
	{
		req->bytes = sendto(network.dhcpConn[req->sockInd].sock,
							req->raw,
							packSize,
							MSG_DONTROUTE,
							(sockaddr*)&req->remote,
							sizeof(req->remote));
	}
	else
	{
		req->bytes = sendto(network.dhcpConn[req->sockInd].sock,
							req->raw,
							packSize,
							0,
							(sockaddr*)&req->remote,
							sizeof(req->remote));
	}

	if (errno || req->bytes <= 0)
		return 0;

	//printf("goes=%s %i\n",IP2String(tempbuff, req->dhcpp.header.bp_yiaddr),req->sockInd);
	return req->dhcpp.header.bp_yiaddr;
}

MYDWORD alad(data9 *req)
{
	//printf("in alad hostname=%s\n", req->hostname);

	if (req->dhcpEntry && (req->req_type == DHCP_MESS_NONE || req->resp_type == DHCP_MESS_ACK))
	{
		MYDWORD retVal = updateDHCP(req);

		if (retVal)
		{
			if (cfig.dhcpLogLevel >= 1)
			{
				if (req->lease && req->reqIP)
				{
					sprintf(logBuff, "Host %s (%s) allotted %s for %u seconds", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_yiaddr), req->lease);
				}
				else if (req->req_type)
				{
					sprintf(logBuff, "Host %s (%s) renewed %s for %u seconds", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_yiaddr), req->lease);
				}
				else
				{
					sprintf(logBuff, "BOOTP Host %s (%s) allotted %s", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_yiaddr));
				}
				logDHCPMess(logBuff, 1);
			}

			if (cfig.replication)
				sendRepl(req);

			return retVal;
		}
	}
	else if (cfig.dhcpLogLevel == 2 && req->resp_type == DHCP_MESS_OFFER)
	{
		sprintf(logBuff, "Host %s (%s) offered %s", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_yiaddr));
		logDHCPMess(logBuff, 2);
	}
	//printf("%u=out\n", req->resp_type);
	return 0;
}

void addOptions(data9 *req)
{
	data3 op;
	int i;

	if (req->req_type && req->resp_type)
	{
		op.opt_code = DHCP_OPTION_MESSAGETYPE;
		op.size = 1;
		op.value[0] = req->resp_type;
		pvdata(req, &op);
	}

	if (req->dhcpEntry && req->resp_type != DHCP_MESS_DECLINE && req->resp_type != DHCP_MESS_NAK)
	{
		strcpy(req->dhcpp.header.bp_sname, cfig.servername);

		if (req->dhcpEntry->fixed)
		{
			//printf("%u,%u\n", req->dhcpEntry->options, *req->dhcpEntry->options);
			MYBYTE *opPointer = req->dhcpEntry->options;

			if (opPointer)
			{
				MYBYTE requestedOnly = *opPointer;
				opPointer++;

				while (*opPointer && *opPointer != DHCP_OPTION_END)
				{
					op.opt_code = *opPointer;
					opPointer++;
					op.size = *opPointer;
					opPointer++;

					if (!requestedOnly || req->paramreqlist[*opPointer])
					{
						memcpy(op.value, opPointer, op.size);
						pvdata(req, &op);
					}
					opPointer += op.size;
				}
			}
		}

		if (req->req_type && req->resp_type)
		{
			if (req->dhcpEntry->rangeInd >= 0)
			{
				MYBYTE *opPointer = cfig.dhcpRanges[req->dhcpEntry->rangeInd].options;
				//printf("Range=%i Pointer=%u\n", req->dhcpEntry->rangeInd,opPointer);

				if (opPointer)
				{
					MYBYTE requestedOnly = *opPointer;
					opPointer++;

					while (*opPointer && *opPointer != DHCP_OPTION_END)
					{
						op.opt_code = *opPointer;
						opPointer++;
						op.size = *opPointer;
						opPointer++;

						if (!requestedOnly || req->paramreqlist[*opPointer])
						{
							memcpy(op.value, opPointer, op.size);
							pvdata(req, &op);
						}
						opPointer += op.size;
					}
				}
			}

			MYBYTE *opPointer = cfig.options;

			if (opPointer)
			{
				MYBYTE requestedOnly = *opPointer;

				opPointer++;
				while (*opPointer && *opPointer != DHCP_OPTION_END)
				{
					op.opt_code = *opPointer;
					opPointer++;
					op.size = *opPointer;
					opPointer++;

					if (!requestedOnly || req->paramreqlist[*opPointer])
					{
						memcpy(op.value, opPointer, op.size);
						pvdata(req, &op);
					}
					opPointer += op.size;
				}
			}

			op.opt_code = DHCP_OPTION_SERVERID;
			op.size = 4;
			pIP(op.value, network.dhcpConn[req->sockInd].server);
			pvdata(req, &op);

			op.opt_code = DHCP_OPTION_DOMAINNAME;
			op.size = strlen(cfig.zone);
			memcpy(op.value, cfig.zone, op.size);
			pvdata(req, &op);

			if (!req->opAdded[DHCP_OPTION_IPADDRLEASE])
			{
				op.opt_code = DHCP_OPTION_IPADDRLEASE;
				op.size = 4;
				pULong(op.value, cfig.lease);
				pvdata(req, &op);
			}

			if (!req->opAdded[DHCP_OPTION_NETMASK])
			{
				op.opt_code = DHCP_OPTION_NETMASK;
				op.size = 4;

				if (req->dhcpEntry->rangeInd >= 0)
					pIP(op.value, cfig.dhcpRanges[req->dhcpEntry->rangeInd].mask);
				else
					pIP(op.value, network.dhcpConn[req->sockInd].mask);

				pvdata(req, &op);
			}

			if (!req->hostname[0])
				genHostName(req->hostname, req->dhcpp.header.bp_chaddr, req->dhcpp.header.bp_hlen);

			if (!req->dhcpEntry->hostname)
					req->dhcpEntry->hostname = cloneString(req->hostname);
			else if (req->hostname[0] && strcasecmp(req->dhcpEntry->hostname, req->hostname))
			{
				free(req->dhcpEntry->hostname);
				req->dhcpEntry->hostname = cloneString(req->hostname);
			}

/*
			if (!req->opAdded[DHCP_OPTION_ROUTER])
			{
				op.opt_code = DHCP_OPTION_ROUTER;
				op.size = 4;
				pIP(op.value, network.dhcpConn[req->sockInd].server);
				pvdata(req, &op);
			}
*/
/*
			if (!req->opAdded[DHCP_OPTION_HOSTNAME])
			{
				op.opt_code = DHCP_OPTION_HOSTNAME;
				op.size = strlen(req->hostname);
				memcpy(op.value, req->hostname, op.size);
				pvdata(req, &op);
			}

			if (req->clientId.opt_code == DHCP_OPTION_CLIENTID)
				pvdata(req, &req->clientId);
*/
			if (req->agentOption.opt_code == DHCP_OPTION_RELAYAGENTINFO)
				pvdata(req, &req->agentOption);
		}
	}

	*(req->vp) = DHCP_OPTION_END;
}

void pvdata(data9 *req, data3 *op)
{
	if (!req->opAdded[op->opt_code] && ((req->vp - (MYBYTE*)&req->dhcpp) + op->size < req->messsize))
	{
		if (op->opt_code == DHCP_OPTION_NEXTSERVER)
			req->dhcpp.header.bp_siaddr = fIP(op->value);
		else if (op->opt_code == DHCP_OPTION_BP_FILE)
		{
			if (op->size <= 128)
				memcpy(req->dhcpp.header.bp_file, op->value, op->size);
		}
		else if(op->size)
		{
			if (op->opt_code == DHCP_OPTION_IPADDRLEASE)
			{
				if (!req->lease || req->lease > fULong(op->value))
					req->lease = fULong(op->value);

				//if (req->dhcpEntry->no_route || req->lease >= INT_MAX)
				if (req->lease >= INT_MAX)
					req->lease = UINT_MAX;

				pULong(op->value, req->lease);
			}
			else if (op->opt_code == DHCP_OPTION_HOSTNAME)
			{
				memcpy(req->hostname, op->value, op->size);
				req->hostname[op->size] = 0;
			}

			memcpy(req->vp, op, (op->size + 2));
			(req->vp) += 2;
			(req->vp) += op->size;
		}
		req->opAdded[op->opt_code] = true;
	}
}

MYDWORD updateDHCP(data9 *req)
{
	data8 *dhcpData = (data8*)calloc(1, sizeof(data8));
	strcpy(dhcpData->hostname, req->hostname);
	memcpy(dhcpData->bp_chaddr, req->dhcpp.header.bp_chaddr, req->dhcpp.header.bp_hlen);
	dhcpData->bp_hlen = req->dhcpp.header.bp_hlen;
	dhcpData->ip = req->dhcpEntry->ip;

	if (!req->req_type || req->lease > (MYDWORD)(INT_MAX - t))
	{
		dhcpData->expiry = INT_MAX;
		req->dhcpEntry->expiry = INT_MAX;
		hangDhcpEntry(req->dhcpEntry);
	}
	else
	{
		dhcpData->expiry = t + req->lease;
		req->dhcpEntry->expiry = dhcpData->expiry;
		hangDhcpEntry(req->dhcpEntry);
	}

	if (req->dhcpEntry->dhcpInd)
		dhcpData->dhcpInd = req->dhcpEntry->dhcpInd;
	else
	{
		cfig.dhcpInd++;
		req->dhcpEntry->dhcpInd = cfig.dhcpInd;
	}

	pthread_t threadId;
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	int errcode = pthread_create(&threadId, &attr, updateStateFile, (void*)dhcpData);
	pthread_attr_destroy(&attr);

	if(errcode)
	{
		if (cfig.dhcpLogLevel >= 1)
		{
			sprintf(logBuff, "Thread Creation Failed");
			logDHCPMess(logBuff, 1);
		}

		free(dhcpData);
	}

	return req->dhcpEntry->ip;
}

void setLeaseExpiry(data7 *dhcpEntry, MYDWORD lease)
{
	if (dhcpEntry && dhcpEntry->ip)
	{
		dhcpEntry->display = false;
		dhcpEntry->hanged = false;

		if (lease > (MYDWORD)(INT_MAX - t))
			dhcpEntry->expiry = INT_MAX;
		else
			dhcpEntry->expiry = t + lease;

		if (dhcpEntry->rangeInd >= 0 && dhcpEntry->rangeInd < MAX_RANGES)
		{
			MYDWORD iip = htonl(dhcpEntry->ip);

			if (iip >= cfig.dhcpRanges[dhcpEntry->rangeInd].rangeStart && iip <= cfig.dhcpRanges[dhcpEntry->rangeInd].rangeEnd)
			{
				int ind = iip - cfig.dhcpRanges[dhcpEntry->rangeInd].rangeStart;

				if (cfig.dhcpRanges[dhcpEntry->rangeInd].expiry[ind] != INT_MAX)
					cfig.dhcpRanges[dhcpEntry->rangeInd].expiry[ind] = dhcpEntry->expiry;

				cfig.dhcpRanges[dhcpEntry->rangeInd].dhcpEntry[ind] = dhcpEntry;
			}
		}
	}
}

void setExtExpiry(data7 *dhcpEntry, MYDWORD lease)
{
	if (dhcpEntry && dhcpEntry->ip)
	{
		dhcpEntry->display = true;
		dhcpEntry->hanged = false;

		if (lease > (MYDWORD)(INT_MAX - t))
			dhcpEntry->expiry = INT_MAX;
		else
			dhcpEntry->expiry = t + lease;

		if (dhcpEntry->rangeInd >= 0 && dhcpEntry->rangeInd < MAX_RANGES)
		{
			MYDWORD iip = htonl(dhcpEntry->ip);

			if (iip >= cfig.dhcpRanges[dhcpEntry->rangeInd].rangeStart && iip <= cfig.dhcpRanges[dhcpEntry->rangeInd].rangeEnd)
			{
				int ind = iip - cfig.dhcpRanges[dhcpEntry->rangeInd].rangeStart;

				if (cfig.dhcpRanges[dhcpEntry->rangeInd].expiry[ind] != INT_MAX)
					cfig.dhcpRanges[dhcpEntry->rangeInd].expiry[ind] = dhcpEntry->expiry;

				cfig.dhcpRanges[dhcpEntry->rangeInd].dhcpEntry[ind] = dhcpEntry;
			}
		}
	}
}

void hangDhcpEntry(data7 *dhcpEntry)
{
	if (dhcpEntry && dhcpEntry->ip)
	{
		dhcpEntry->display = true;

		if (!dhcpEntry->hanged && dhcpEntry->expiry > t)
		{
//			dhcpAge.insert(pair<time_t, data7*>(dhcpEntry->expiry, dhcpEntry));
			dhcpEntry->hanged = true;
		}

		if (dhcpEntry->rangeInd >= 0 && dhcpEntry->rangeInd < MAX_RANGES)
		{
			MYDWORD iip = htonl(dhcpEntry->ip);

			if (iip >= cfig.dhcpRanges[dhcpEntry->rangeInd].rangeStart && iip <= cfig.dhcpRanges[dhcpEntry->rangeInd].rangeEnd)
			{
				int ind = iip - cfig.dhcpRanges[dhcpEntry->rangeInd].rangeStart;

				if (cfig.dhcpRanges[dhcpEntry->rangeInd].expiry[ind] != INT_MAX)
					cfig.dhcpRanges[dhcpEntry->rangeInd].expiry[ind] = dhcpEntry->expiry;

				cfig.dhcpRanges[dhcpEntry->rangeInd].dhcpEntry[ind] = dhcpEntry;
			}
		}
	}
}

void lockIP(MYDWORD ip)
{
	if (ip)
	{
		MYDWORD iip = htonl(ip);

		for (char rangeInd = 0; rangeInd < MAX_RANGES && cfig.dhcpRanges[rangeInd].rangeStart; rangeInd++)
		{
			if (iip >= cfig.dhcpRanges[rangeInd].rangeStart && iip <= cfig.dhcpRanges[rangeInd].rangeEnd)
			{
				int ind = iip - cfig.dhcpRanges[rangeInd].rangeStart;
				cfig.dhcpRanges[rangeInd].expiry[ind] = INT_MAX;
				break;
			}
		}
	}
}

void *sendToken(void *lparam)
{
	while (true)
	{
		token.bytes = token.vp - (MYBYTE*)token.raw;

		sendto(cfig.dhcpReplConn.sock,
				token.raw,
				token.bytes,
				0,
				(sockaddr*)&token.remote,
				sizeof(token.remote));

		sleep(500);
	}

	pthread_exit(NULL);
}

MYDWORD sendRepl(data9 *req)
{
	data3 op;

	if (cfig.replication == 1)
		req->remote.sin_addr.s_addr = cfig.zoneServers[1];
	else if (cfig.replication == 2)
		req->remote.sin_addr.s_addr = cfig.zoneServers[0];
	else
		return 0;

	req->remote.sin_port = htons(IPPORT_DHCPS);
	req->remote.sin_family = AF_INET;

	MYBYTE *opPointer = req->dhcpp.vend_data;

	while ((*opPointer) != DHCP_OPTION_END && opPointer < req->vp)
	{
		if ((*opPointer) == DHCP_OPTION_MESSAGETYPE)
		{
			*(opPointer + 2) = DHCP_MESS_INFORM;
			break;
		}
		opPointer = opPointer + *(opPointer + 1) + 2;
	}

	if (!req->opAdded[DHCP_OPTION_MESSAGETYPE])
	{
		op.opt_code = DHCP_OPTION_MESSAGETYPE;
		op.size = 1;
		op.value[0] = DHCP_MESS_INFORM;
		pvdata(req, &op);
	}

	if (req->hostname[0] && !req->opAdded[DHCP_OPTION_HOSTNAME])
	{
		op.opt_code = DHCP_OPTION_HOSTNAME;
		op.size = strlen(req->hostname);
		memcpy(op.value, req->hostname, op.size);
		pvdata(req, &op);
	}

	if (!req->opAdded[DHCP_OPTION_IPADDRLEASE])
	{
		op.opt_code = DHCP_OPTION_IPADDRLEASE;
		op.size = 4;
		pULong(op.value, req->lease);
		pvdata(req, &op);
	}

//	op.opt_code = DHCP_OPTION_SERIAL;
//	op.size = 4;
//	pULong(op.value, cfig.serial1);
//	pvdata(req, &op);

	*(req->vp) = DHCP_OPTION_END;
	req->vp++;
	req->bytes = req->vp - (MYBYTE*)req->raw;

	req->dhcpp.header.bp_op = BOOTP_REQUEST;
	errno = 0;

	sendto(cfig.dhcpReplConn.sock,
			req->raw,
			req->bytes,
			0,
			(sockaddr*)&req->remote,
			sizeof(req->remote));

	//errno = WSAGetLastError();

	if (errno || req->bytes <= 0)
	{
		if (cfig.dhcpLogLevel >= 1)
		{
			if (cfig.replication == 1)
				sprintf(logBuff, "Error %s Sending DHCP Update to Secondary Server", strerror(errno));
			else
				sprintf(logBuff, "Error %s Sending DHCP Update to Primary Server", strerror(errno));

			logDHCPMess(logBuff, 1);
		}

		return 0;
	}
	else if (cfig.dhcpLogLevel == 2)
	{
		if (cfig.replication == 1)
			sprintf(logBuff, "DHCP Update for host %s (%s) sent to Secondary Server", req->dhcpEntry->mapname, IP2String(tempbuff, req->dhcpEntry->ip));
		else
			sprintf(logBuff, "DHCP Update for host %s (%s) sent to Primary Server", req->dhcpEntry->mapname, IP2String(tempbuff, req->dhcpEntry->ip));

		logDHCPMess(logBuff, 2);
	}

	return req->dhcpp.header.bp_yiaddr;
}

MYDWORD sendRepl(data7 *dhcpEntry)
{
	data9 req;
	memset(&req, 0, sizeof(data9));
	req.vp = req.dhcpp.vend_data;
	req.messsize = sizeof(dhcp_packet);
	req.dhcpEntry = dhcpEntry;

	req.dhcpp.header.bp_op = BOOTP_REQUEST;
	req.dhcpp.header.bp_xid = t;
	req.dhcpp.header.bp_ciaddr = dhcpEntry->ip;
	req.dhcpp.header.bp_yiaddr = dhcpEntry->ip;
	req.dhcpp.header.bp_hlen = 16;
	getHexValue(req.dhcpp.header.bp_chaddr, req.dhcpEntry->mapname, &(req.dhcpp.header.bp_hlen));
	req.dhcpp.header.bp_magic_num[0] = 99;
	req.dhcpp.header.bp_magic_num[1] = 130;
	req.dhcpp.header.bp_magic_num[2] = 83;
	req.dhcpp.header.bp_magic_num[3] = 99;
	strcpy(req.hostname, dhcpEntry->hostname);

	return sendRepl(&req);
}

void recvRepl(data9 *req)
{
	MYDWORD ip = req->dhcpp.header.bp_yiaddr ? req->dhcpp.header.bp_yiaddr : req->dhcpp.header.bp_ciaddr;

	if (!ip || !req->dhcpp.header.bp_hlen)
		return;

	char rInd = getRangeInd(ip);

	if (rInd >= 0)
	{
		int ind  = getIndex(rInd, ip);
		data7 *dhcpEntry = cfig.dhcpRanges[rInd].dhcpEntry[ind];

		if (dhcpEntry && ip == dhcpEntry->ip && dhcpEntry->expiry > t && strcasecmp(dhcpEntry->mapname, req->chaddr))
		{
			if (cfig.replication == 1)
				sprintf(logBuff, "DHCP Update ignored for %s (%s) from Secondary Server", req->chaddr, IP2String(tempbuff, ip));
			else
				sprintf(logBuff, "DHCP Update ignored for %s (%s) from Primary Server", req->chaddr, IP2String(tempbuff, ip));

			logDHCPMess(logBuff, 1);
			return;
		}
	}

	req->dhcpEntry = findDHCPEntry(req->chaddr);

	if (!req->dhcpEntry && rInd >= 0)
	{
		req->dhcpEntry = (data7*)calloc(1, sizeof(data7));

		if (!req->dhcpEntry)
		{
			sprintf(logBuff, "Memory Allocation Error");
			logDHCPMess(logBuff, 1);
			return;
		}

		req->dhcpEntry->mapname = cloneString(req->chaddr);

		if (!req->dhcpEntry->mapname)
		{
			sprintf(logBuff, "Memory Allocation Error");
			free(req->dhcpEntry);
			logDHCPMess(logBuff, 1);
			return;
		}

		dhcpCache[req->dhcpEntry->mapname] = req->dhcpEntry;
	}

	if (req->dhcpEntry && (rInd >=0 || (req->dhcpEntry->fixed && req->dhcpEntry->ip == ip)))
	{
		req->dhcpEntry->ip = ip;
		req->dhcpEntry->rangeInd = rInd;

		if (!req->dhcpEntry->hostname)
				req->dhcpEntry->hostname = cloneString(req->hostname);
		else if (req->hostname[0] && strcasecmp(req->dhcpEntry->hostname, req->hostname))
		{
			free(req->dhcpEntry->hostname);
			req->dhcpEntry->hostname = cloneString(req->hostname);
		}

		if (!req->req_type || req->lease > (MYDWORD)(INT_MAX - t))
			setExtExpiry(req->dhcpEntry, INT_MAX);
		else
			setExtExpiry(req->dhcpEntry, req->lease);
	}
	else
	{
		if (cfig.replication == 1)
			sprintf(logBuff, "DHCP Update ignored for %s (%s) from Secondary Server", req->chaddr, IP2String(tempbuff, ip));
		else
			sprintf(logBuff, "DHCP Update ignored for %s (%s) from Primary Server", req->chaddr, IP2String(tempbuff, ip));

		logDHCPMess(logBuff, 1);
		return;
	}

	if (cfig.dhcpLogLevel == 2)
	{
		if (cfig.replication == 1)
			sprintf(logBuff, "DHCP Update received for %s (%s) from Secondary Server", req->chaddr, IP2String(tempbuff, ip));
		else
			sprintf(logBuff, "DHCP Update received for %s (%s) from Primary Server", req->chaddr, IP2String(tempbuff, ip));

		logDHCPMess(logBuff, 2);
	}
}

data7 *findDHCPEntry(char *key)
{
	//printf("finding %u=%s\n",ind,key);
	myLower(key);
	dhcpMap::iterator it = dhcpCache.find(key);

	if (it == dhcpCache.end())
		return NULL;
	else
		return it->second;
}

char getRangeInd(MYDWORD ip)
{
	if (ip)
	{
		MYDWORD iip = htonl(ip);
		for (char k = 0; k < MAX_RANGES && cfig.dhcpRanges[k].rangeStart; k++)
			if (iip >= cfig.dhcpRanges[k].rangeStart && iip <= cfig.dhcpRanges[k].rangeEnd)
				return k;
	}
	return -1;
}

int getIndex(char rangeInd, MYDWORD ip)
{
	if (ip && rangeInd >= 0 && rangeInd < MAX_RANGES)
	{
		MYDWORD iip = htonl(ip);
		if (iip >= cfig.dhcpRanges[rangeInd].rangeStart && iip <= cfig.dhcpRanges[rangeInd].rangeEnd)
			return (iip - cfig.dhcpRanges[rangeInd].rangeStart);
	}
	return -1;
}

void loadOptions(FILE *f, const char *sectionName, data20 *optionData)
{
	optionData->options = NULL;
	optionData->ip = 0;
	optionData->mask = 0;
	optionData->lease = 0;
	MYBYTE maxInd = sizeof(opData) / sizeof(data4);
	MYBYTE options[sizeof(dhcp_packet) - sizeof(dhcp_header)];
	MYWORD buffsize = sizeof(options);
	MYBYTE *dp = options;
	MYBYTE op_specified[256];

	memset(op_specified, 0, 256);
	*dp = 0;
	dp++;

	char raw[512];
	char name[512];
	char value[512];

	while (readSection(raw, f))
	{
		MYBYTE *ddp = dp;
		MYBYTE hoption[256];
		MYBYTE valSize = sizeof(hoption) - 1;
		MYBYTE opTag = 0;
		MYBYTE opType = 0;
		MYBYTE valType = 0;
		bool tagFound = false;

		mySplit(name, value, raw, '=');

		if (!name[0])
		{
			sprintf(logBuff, "Warning: section [%s] invalid option %s ignored", sectionName, raw);
			logDHCPMess(logBuff, 1);
			continue;
		}

		if (!strcasecmp(name, "DHCPRange"))
		{
			if (!strcasecmp(sectionName, RANGESET))
				addDHCPRange(value);
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s not allowed in this section, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
			}
			continue;
		}
		else if (!strcasecmp(name, "IP"))
		{
			if (!strcasecmp(sectionName, GLOBALOPTIONS) || !strcasecmp(sectionName, RANGESET))
			{
				sprintf(logBuff, "Warning: section [%s] option %s not allowed in this section, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
			}
			else if (!isIP(value))
			{
				sprintf(logBuff, "Warning: section [%s] option %s Invalid IP Addr %s option ignored", sectionName, value);
				logDHCPMess(logBuff, 1);
			}
			else
				optionData->ip = inet_addr(value);

			continue;
		}
		else if (!strcasecmp(name, "FilterMacRange"))
		{
			if (!strcasecmp(sectionName, RANGESET))
				addMacRange(optionData->rangeSetInd, value);
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s not allowed in this section, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
			}
			continue;
		}

		if (isInt(name))
		{
			if (atoi(name) >= 254)
			{
				sprintf(logBuff, "Warning: section [%s] invalid option %s, ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
				continue;
			}

			opTag = atoi(name);
			opType = 0;
		}

		for (MYBYTE i = 0; i < maxInd; i++)
			if (!strcasecmp(name, opData[i].opName) || (opTag && opTag == opData[i].opTag))
			{
				opTag = opData[i].opTag;
				opType = opData[i].opType;
				tagFound = true;
				break;
			}

		if (!value[0])
			valType = 9;
		else if (value[0] == '"' && value[strlen(value)-1] == '"')
		{
			valType = 1;
			value[0] = 32;
			value[strlen(value)-1] = 32;
			myTrim(value, value);

			if (strlen(value) <= UCHAR_MAX)
				valSize = strlen(value);
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s value too big, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
				continue;
			}
		}
		else if (strchr(value, ':'))
		{
			valType = 2;
			valSize = sizeof(hoption) - 1;
			char *errorPos = getHexValue(hoption, value, &valSize);

			if (errorPos)
			{
				sprintf(logBuff, "Warning: section [%s] option %s position %u, Invalid char %c, option ignored", sectionName, name, (errorPos - value) + 1, *(errorPos));
				logDHCPMess(logBuff, 1);
				continue;
			}
			else
				memcpy(value, hoption, valSize);
		}
		else if (isInt(value) && atol(value) > USHRT_MAX)
			valType = 4;
		else if (isInt(value) && atoi(value) > UCHAR_MAX)
			valType = 5;
		else if (isInt(value))
			valType = 6;
		else if ((strchr(value, '.') && (opType == 2 || opType == 3 || opType == 8)) || (!strchr(value, '.') && strchr(value, ',')))
		{
			valType = 2;
			myTokenize(tempbuff, value, "/,.", true);
			char *ptr = tempbuff;
			valSize = 0;

			for (; *ptr; ptr = myGetToken(ptr, 1))
			{
				//printf("%s:", ptr);
				if (valSize >= UCHAR_MAX)
				{
					sprintf(logBuff, "Warning: section [%s] option %s, too many values, entry ignored", sectionName, raw);
					logDHCPMess(logBuff, 1);
					break;
				}
				else if (isInt(ptr) && atoi(ptr) <= UCHAR_MAX)
				{
					hoption[valSize] = atoi(ptr);
					valSize++;
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s invalid byte/octate '%s', entry ignored", sectionName, name, ptr);
					logDHCPMess(logBuff, 1);
					break;
				}
			}

			if (*ptr)
				continue;

			memcpy(value, hoption, valSize);
		}
		else
		{
			if (strlen(value) <= UCHAR_MAX)
				valSize = strlen(value);
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s value too long, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
				continue;
			}
			valType = 1;
		}

		if (!strcasecmp(name, "FilterVendorClass"))
		{
			if (!strcasecmp(sectionName, RANGESET))
				addVendClass(optionData->rangeSetInd, value, valSize);
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s not allowed in this section, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
			}
			continue;
		}
		else if (!strcasecmp(name, "FilterUserClass"))
		{
			if (!strcasecmp(sectionName, RANGESET))
				addUserClass(optionData->rangeSetInd, value, valSize);
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s not allowed in this section, option ignored", sectionName, raw);
				logDHCPMess(logBuff, 1);
			}
			continue;
		}

		if (!opTag)
		{
			sprintf(logBuff, "Warning: section [%s] invalid option %s, ignored", sectionName, raw);
			logDHCPMess(logBuff, 1);
			continue;
		}

		if (!opType)
			opType = valType;

		//sprintf(logBuff, "Tag %i ValType %i opType %i value=%s size=%u", opTag, valType, opType, value, valSize);
		//logDHCPMess(logBuff, 1);

		if (op_specified[opTag])
		{
			sprintf(logBuff, "Warning: section [%s] duplicate option %s, ignored", sectionName, raw);
			logDHCPMess(logBuff, 1);
			continue;
		}

		//printf("Option=%u opType=%u valueType=%u valSize=%u\n", opTag, opType, valType, valSize);

		op_specified[opTag] = true;

		if (valType == 9)
		{
			if (buffsize > 2)
			{
				*dp = opTag;
				dp++;
				*dp = 0;
				dp++;
				buffsize -= 2;
			}
			else
			{
				sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
				logDHCPMess(logBuff, 1);
			}
			continue;
		}

		switch (opType)
		{
			case 1:
			{
				if (valType != 1 && valType != 2)
				{
					sprintf(logBuff, "Warning: section [%s] option %s, need string value, option ignored", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
//				else if (opTag == DHCP_OPTION_DOMAINNAME)
//				{
//					sprintf(logBuff, "Warning: section [%s] option %u should be under [DOMAIN_NAME], ignored", sectionName, opTag);
//					logDHCPMess(logBuff, 1);
//					continue;
//				}
				else if (buffsize > strlen(value) + 2)
				{
					*dp = opTag;
					dp++;
					*dp = strlen(value);
					dp++;
					memcpy(dp, value, valSize);
					dp += (strlen(value));
					buffsize -= (valSize + 2);
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;

			case 3:
			case 8:
			{
				if (valType == 2)
				{
					if (opType == 3 && valSize % 4)
					{
						sprintf(logBuff, "Warning: section [%s] option %s, missing bytes of IP, option ignored", sectionName, raw);
						logDHCPMess(logBuff, 1);
						continue;
					}
					else if (opType == 8 && valSize % 8)
					{
						sprintf(logBuff, "Warning: section [%s] option %s, some values not in IP/Mask form, option ignored", sectionName, raw);
						logDHCPMess(logBuff, 1);
						continue;
					}

					if (opTag == DHCP_OPTION_NETMASK)
					{
						if (valSize != 4 || !checkMask(*((MYDWORD*)value)))
						{
							sprintf(logBuff, "Warning: section [%s] Invalid subnetmask %s, option ignored", sectionName, raw);
							logDHCPMess(logBuff, 1);
							continue;
						}
						else
							optionData->mask = *((MYDWORD*)value);
					}

					if (buffsize > valSize + 2)
					{
						*dp = opTag;
						dp++;
						*dp = valSize;
						dp++;
						memcpy(dp, value, valSize);
						dp += valSize;
						buffsize -= (valSize + 2);
					}
					else
					{
						sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
						logDHCPMess(logBuff, 1);
					}
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, Invalid value, should be one or more IP/4 Bytes", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;

			case 4:
			{
				MYDWORD j;

				if (valType == 2 && valSize == 4)
					j = fULong(value);
				else if (valType >= 4 && valType <= 6)
					j = atol(value);
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, value should be integer between 0 & %u or 4 bytes, option ignored", sectionName, name, UINT_MAX);
					logDHCPMess(logBuff, 1);
					continue;
				}

				if (opTag == DHCP_OPTION_IPADDRLEASE)
				{
					if (j == 0)
						j = UINT_MAX;

//					if (!strcasecmp(sectionName, GLOBALOPTIONS))
//					{
//						sprintf(logBuff, "Warning: section [%s] option %s not allowed in this section, please set it in [TIMINGS] section", sectionName, raw);
//						logDHCPMess(logBuff, 1);
//						continue;
//					}

					optionData->lease = j;
				}

				if (buffsize > 6)
				{
					*dp = opTag;
					dp++;
					*dp = 4;
					dp++;
					dp += pULong(dp, j);
					buffsize -= 6;
					//printf("%s=%u=%u\n",opData[op_index].opName,opData[op_index].opType,htonl(j));
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;

			case 5:
			{
				MYWORD j;

				if (valType == 2 && valSize == 2)
					j = fUShort(value);
				else if (valType == 5 || valType == 6)
					j = atol(value);
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, value should be between 0 & %u or 2 bytes, option ignored", sectionName, name, USHRT_MAX);
					logDHCPMess(logBuff, 1);
					continue;
				}

				if (buffsize > 4)
				{
					*dp = opTag;
					dp++;
					*dp = 2;
					dp++;
					dp += pUShort(dp, j);
					buffsize -= 4;
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;

			case 6:
			{
				MYBYTE j;

				if (valType == 2 && valSize == 1)
					j = *value;
				else if (valType == 6)
					j = atol(value);
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, value should be between 0 & %u or single byte, option ignored", sectionName, name, UCHAR_MAX);
					logDHCPMess(logBuff, 1);
					continue;
				}

				if (buffsize > 3)
				{
					*dp = opTag;
					dp++;
					*dp = 1;
					dp++;
					*dp = j;
					dp++;
					buffsize -= 3;
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;

			case 7:
			{
				MYBYTE j;

				if (valType == 2 && valSize == 1 && *value < 2)
					j = *value;
				else if (valType == 1 && !strcasecmp(value, "yes"))
					j = 1;
				else if (valType == 1 && !strcasecmp(value, "no"))
					j = 0;
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, value should be yes or no, option ignored", sectionName, raw);
					logDHCPMess(logBuff, 1);
					continue;
				}

				if (buffsize > 3)
				{
					*dp = opTag;
					dp++;
					*dp = 1;
					dp++;
					*dp = j;
					dp++;
					buffsize -= 3;
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;

			default:
			{
				if (valType == 6)
				{
					valType = 2;
					valSize = 1;
					*value = atoi(value);
				}

				if (opType == 2 && valType != 2)
				{
					sprintf(logBuff, "Warning: section [%s] option %s, value should be comma separated bytes or hex string, option ignored", sectionName, raw);
					logDHCPMess(logBuff, 1);
					continue;
				}
				else if (buffsize > valSize + 2)
				{
					*dp = opTag;
					dp++;
					*dp = valSize;
					dp++;
					memcpy(dp, value, valSize);
					dp += valSize;
					buffsize -= (valSize + 2);
				}
				else
				{
					sprintf(logBuff, "Warning: section [%s] option %s, no more space for options", sectionName, raw);
					logDHCPMess(logBuff, 1);
				}
			}
			break;
		}

		//printf("%s Option=%u opType=%u valType=%u  valSize=%u\n", raw, opTag, opType, valType, valSize);
		//printf("%s %s\n", name, hex2String(tempbuff, ddp, valSize+2, ':'));
	}

	//printf("%s=%s\n", sectionName, optionData->vendClass);

	*dp = DHCP_OPTION_END;
	dp++;
	MYWORD optionSize = (dp - options);

	if (optionSize > 2)
	{
		optionData->options = (MYBYTE*)calloc(1, optionSize);

		if (!optionData->options)
		{
			sprintf(logBuff, "DHCP Option Load, Memory Allocation Error");
			logDHCPMess(logBuff, 1);
		}

		memcpy(optionData->options, options, optionSize);
	}
}

void lockOptions(FILE *f)
{
	char raw[512];
	char name[512];
	char value[512];

	while (readSection(raw, f))
	{
		mySplit(name, value, raw, '=');

		if (!name[0] || !value[0])
			continue;

		int op_index;
		MYBYTE n = sizeof(opData) / sizeof(data4);

		for (op_index = 0; op_index < n; op_index++)
			if (!strcasecmp(name, opData[op_index].opName) || (opData[op_index].opTag && atoi(name) == opData[op_index].opTag))
				break;

		if (op_index >= n)
			continue;

		if (opData[op_index].opType == 3)
		{
			if (myTokenize(value, value, "/,.", true))
			{
				char *ptr = value;
				char hoption[256];
				MYBYTE valueSize = 0;

				for (; *ptr; ptr = myGetToken(ptr, 1))
				{
					if (valueSize >= UCHAR_MAX)
						break;
					else if (isInt(ptr) && atoi(ptr) <= UCHAR_MAX)
					{
						hoption[valueSize] = atoi(ptr);
						valueSize++;
					}
					else
						break;
				}

				if (*ptr)
					continue;

				if (valueSize % 4)
					continue;

				for (MYBYTE i = 0; i < valueSize; i += 4)
				{
					MYDWORD inetAddr = *((MYDWORD*)&(hoption[i]));

					if (inetAddr != INADDR_ANY && inetAddr != INADDR_NONE)
						lockIP(inetAddr);
				}
			}
		}
	}
}

void addDHCPRange(char *dp)
{
	MYDWORD rs = 0;
	MYDWORD re = 0;
	char name[512];
	char value[512];
	mySplit(name, value, dp, '-');

	if (isIP(name) && isIP(value))
	{
		rs = htonl(inet_addr(name));
		re = htonl(inet_addr(value));

		if (rs && re && rs <= re)
		{
			data13 *range;
			MYBYTE m = 0;

			for (; m < MAX_RANGES && cfig.dhcpRanges[m].rangeStart; m++)
			{
				range = &cfig.dhcpRanges[m];
				if ((rs >= range->rangeStart && rs <= range->rangeEnd)
						|| (re >= range->rangeStart && re <= range->rangeEnd)
						|| (range->rangeStart >= rs && range->rangeStart <= re)
						|| (range->rangeEnd >= rs && range->rangeEnd <= re))
				{
					sprintf(logBuff, "Warning: DHCP Range %s overlaps with another range, ignored", dp);
					logDHCPMess(logBuff, 1);
					return;
				}
			}

			if (m < MAX_RANGES)
			{
				cfig.dhcpSize += (re - rs + 1);
				range = &cfig.dhcpRanges[m];
				range->rangeStart = rs;
				range->rangeEnd = re;
				range->expiry = (time_t*)calloc((re - rs + 1), sizeof(time_t));
				range->dhcpEntry = (data7**)calloc((re - rs + 1), sizeof(data7*));

				if (!range->expiry || !range->dhcpEntry)
				{
					if (range->expiry)
						free(range->expiry);

					if (range->dhcpEntry)
						free(range->dhcpEntry);

					sprintf(logBuff, "DHCP Ranges Load, Memory Allocation Error");
					logDHCPMess(logBuff, 1);
					return;
				}
			}
		}
		else
		{
			sprintf(logBuff, "Section [%s] Invalid DHCP range %s in ini file, ignored", RANGESET, dp);
			logDHCPMess(logBuff, 1);
		}
	}
	else
	{
		sprintf(logBuff, "Section [%s] Invalid DHCP range %s in ini file, ignored", RANGESET, dp);
		logDHCPMess(logBuff, 1);
	}
}

void addVendClass(MYBYTE rangeSetInd, char *vendClass, MYBYTE vendClassSize)
{
	data14 *rangeSet = &cfig.rangeSet[rangeSetInd];

	MYBYTE i = 0;

	for (; i <= 32 && rangeSet->vendClassSize[i]; i++);

	if (i >= 32 || !vendClassSize)
		return;

	rangeSet->vendClass[i] = (MYBYTE*)calloc(vendClassSize, 1);

	if(!rangeSet->vendClass[i])
	{
		sprintf(logBuff, "Vendor Class Load, Memory Allocation Error");
		logDHCPMess(logBuff, 1);
	}
	else
	{
		cfig.hasFilter = true;
		rangeSet->vendClassSize[i] = vendClassSize;
		memcpy(rangeSet->vendClass[i], vendClass, vendClassSize);
		//printf("Loaded Vendor Class %s Size=%i rangeSetInd=%i Ind=%i\n", rangeSet->vendClass[i], rangeSet->vendClassSize[i], rangeSetInd, i);
		//printf("Loaded Vendor Class %s Size=%i rangeSetInd=%i Ind=%i\n", hex2String(tempbuff, rangeSet->vendClass[i], rangeSet->vendClassSize[i], ':'), rangeSet->vendClassSize[i], rangeSetInd, i);
	}
}

void addUserClass(MYBYTE rangeSetInd, char *userClass, MYBYTE userClassSize)
{
	data14 *rangeSet = &cfig.rangeSet[rangeSetInd];

	MYBYTE i = 0;

	for (; i <= 32 && rangeSet->userClassSize[i]; i++);

	if (i >= 32 || !userClassSize)
		return;

	rangeSet->userClass[i] = (MYBYTE*)calloc(userClassSize, 1);

	if(!rangeSet->userClass[i])
	{
		sprintf(logBuff, "Vendor Class Load, Memory Allocation Error");
		logDHCPMess(logBuff, 1);
	}
	else
	{
		cfig.hasFilter = true;
		rangeSet->userClassSize[i] = userClassSize;
		memcpy(rangeSet->userClass[i], userClass, userClassSize);
		//printf("Loaded User Class %s Size=%i rangeSetInd=%i Ind=%i\n", hex2String(tempbuff, rangeSet->userClass[i], rangeSet->userClassSize[i], ':'), rangeSet->vendClassSize[i], rangeSetInd, i);
	}
}

void addMacRange(MYBYTE rangeSetInd, char *macRange)
{
	if (macRange[0])
	{
		data14 *rangeSet = &cfig.rangeSet[rangeSetInd];

		MYBYTE i = 0;

		for (; i <= 32 && rangeSet->macSize[i]; i++);

		if (i >= 32)
			return;

		char name[256];
		char value[256];

		mySplit(name, value, macRange, '-');

		if(!name[0] || !value[0])
		{
			sprintf(logBuff, "Section [%s], invalid Filter_Mac_Range %s, ignored", RANGESET, macRange);
			logDHCPMess(logBuff, 1);
		}
		else
		{
			MYBYTE macSize1 = 16;
			MYBYTE macSize2 = 16;
			MYBYTE *macStart = (MYBYTE*)calloc(1, macSize1);
			MYBYTE *macEnd = (MYBYTE*)calloc(1, macSize2);

			if(!macStart || !macEnd)
			{
				sprintf(logBuff, "DHCP Range Load, Memory Allocation Error");
				logDHCPMess(logBuff, 1);
			}
			else if (getHexValue(macStart, name, &macSize1) || getHexValue(macEnd, value, &macSize2))
			{
				sprintf(logBuff, "Section [%s], Invalid character in Filter_Mac_Range %s", RANGESET, macRange);
				logDHCPMess(logBuff, 1);
				free(macStart);
				free(macEnd);
			}
			else if (memcmp(macStart, macEnd, 16) > 0)
			{
				sprintf(logBuff, "Section [%s], Invalid Filter_Mac_Range %s, (higher bound specified on left), ignored", RANGESET, macRange);
				logDHCPMess(logBuff, 1);
				free(macStart);
				free(macEnd);
			}
			else if (macSize1 != macSize2)
			{
				sprintf(logBuff, "Section [%s], Invalid Filter_Mac_Range %s, (start/end size mismatched), ignored", RANGESET, macRange);
				logDHCPMess(logBuff, 1);
				free(macStart);
				free(macEnd);
			}
			else
			{
				cfig.hasFilter = true;
				rangeSet->macSize[i] = macSize1;
				rangeSet->macStart[i] = macStart;
				rangeSet->macEnd[i] = macEnd;
				//printf("Mac Loaded, Size=%i Start=%s rangeSetInd=%i Ind=%i\n", rangeSet->macSize[i], hex2String(tempbuff, rangeSet->macStart[i], rangeSet->macSize[i], ':'), rangeSetInd, i);
			}
		}
	}
}

void loadDHCP()
{
	data7 *dhcpEntry;
	char mapname[64];
	FILE *f = NULL;

	if (f = openSection(GLOBALOPTIONS, 1, iniFile))
	{
		data20 optionData;
		loadOptions(f, GLOBALOPTIONS, &optionData);
		cfig.options = optionData.options;
		cfig.mask = optionData.mask;
		cfig.lease = optionData.lease;
	}

	if (!cfig.lease)
		cfig.lease = 36000;

	if (!cfig.mask)
		cfig.mask = inet_addr("255.255.255.0");

	for (MYBYTE i = 1; i <= MAX_RANGES ; i++)
	{
		if (f = openSection(RANGESET, i, iniFile))
		{
			MYBYTE m = 0;
			for (; m < MAX_RANGES && cfig.dhcpRanges[m].rangeStart; m++);

			data20 optionData;
			optionData.rangeSetInd = i - 1;
			loadOptions(f, RANGESET, &optionData);
			cfig.rangeSet[optionData.rangeSetInd].active = true;

			for (; m < MAX_RANGES && cfig.dhcpRanges[m].rangeStart; m++)
			{
				//printf("Range Start=%s rangeSetInd=%i\n", IP2String(tempbuff, htonl(cfig.dhcpRanges[m].rangeStart)), optionData.rangeSetInd);
				cfig.dhcpRanges[m].rangeSetInd = optionData.rangeSetInd;
				cfig.dhcpRanges[m].options = optionData.options;
				cfig.dhcpRanges[m].mask = optionData.mask;
			}
		}
		else
			break;
	}

	//printf("%s\n", IP2String(tempbuff, cfig.mask));

	for (char rangeInd = 0; rangeInd < MAX_RANGES && cfig.dhcpRanges[rangeInd].rangeStart; rangeInd++)
	{
		if (!cfig.dhcpRanges[rangeInd].mask)
			cfig.dhcpRanges[rangeInd].mask = cfig.mask;

		for (MYDWORD iip = cfig.dhcpRanges[rangeInd].rangeStart; iip <= cfig.dhcpRanges[rangeInd].rangeEnd; iip++)
		{
			MYDWORD ip = htonl(iip);

			if ((cfig.dhcpRanges[rangeInd].mask | (~ip)) == UINT_MAX || (cfig.dhcpRanges[rangeInd].mask | ip) == UINT_MAX)
				cfig.dhcpRanges[rangeInd].expiry[iip - cfig.dhcpRanges[rangeInd].rangeStart] = INT_MAX;
		}
	}

	if (FILE *f = openSection(GLOBALOPTIONS, 1, iniFile))
		lockOptions(f);

	for (MYBYTE i = 1; i <= 32 ;i++)
	{
		if (FILE *f = openSection(RANGESET, i, iniFile))
			lockOptions(f);
		else
			break;
	}

	f = fopen(iniFile, "rt");

	if (f)
	{
		char sectionName[512];

		while (fgets(sectionName, 511, f))
		{
			if (*sectionName == '[')
			{
				char *secend = strchr(sectionName, ']');

				if (secend)
				{
					*secend = 0;
					sectionName[0] = 32;
					myTrim(sectionName, sectionName);
				}
				else
					continue;
			}
			else
				continue;


			MYBYTE hexValue[UCHAR_MAX];
			MYBYTE hexValueSize = sizeof(hexValue);
			data20 optionData;

			if (!getHexValue(hexValue, sectionName, &hexValueSize))
			{
				dhcpEntry = findDHCPEntry(hex2String(mapname, hexValue, hexValueSize));

				if (!dhcpEntry)
				{
					FILE *ff = openSection(sectionName, 1, iniFile);
					loadOptions(ff, sectionName, &optionData);
					ff = openSection(sectionName, 1, iniFile);
					lockOptions(ff);

					if (optionData.ip)
					{
						dhcpMap::iterator p = dhcpCache.begin();

						for (; p != dhcpCache.end(); p++)
						{
							if (p->second && p->second->ip == optionData.ip)
								break;
						}

						if (p == dhcpCache.end())
						{
							dhcpEntry = (data7*)calloc(1, sizeof(data7));

							if (!dhcpEntry)
							{
								sprintf(logBuff, "Host Options Load, Memory Allocation Error");
								logDHCPMess(logBuff, 1);
								return;
							}

							dhcpEntry->mapname = cloneString(mapname);

							if (!dhcpEntry->mapname)
							{
								sprintf(logBuff, "Host Data Load, Memory Allocation Error");
								logDHCPMess(logBuff, 1);
								return;
							}

							dhcpEntry->ip = optionData.ip;
							dhcpEntry->options = optionData.options;
							dhcpEntry->rangeInd = getRangeInd(optionData.ip);
							dhcpEntry->fixed = 1;
							lockIP(optionData.ip);
							dhcpCache[dhcpEntry->mapname] = dhcpEntry;
							//printf("%s=%s=%s\n", mapname, dhcpEntry->mapname, IP2String(tempbuff, optionData.ip));
						}
						else
						{
							sprintf(logBuff, "Static DHCP Host [%s] Duplicate IP Address %s, Entry ignored", sectionName, IP2String(tempbuff, optionData.ip));
							logDHCPMess(logBuff, 1);
						}
					}
					else
					{
						sprintf(logBuff, "IP Address is missing in Static DHCP Host [%s], Entry ignored", sectionName);
						logDHCPMess(logBuff, 1);
					}
				}
				else
				{
					sprintf(logBuff, "Duplicate Static DHCP Host [%s] ignored", sectionName);
					logDHCPMess(logBuff, 1);
				}
			}
			else if (strchr(sectionName, ':'))
			{
				sprintf(logBuff, "Invalid Static DHCP Host MAC Addr [%s] ignored", sectionName);
				logDHCPMess(logBuff, 1);
			}
		}

		fclose(f);
	}

	f = fopen(leaFile, "rb");

	if (f)
	{
		data8 dhcpData;

		while (fread(&dhcpData, sizeof(data8), 1, f))
		{
			char rangeInd = -1;
			int ind = -1;

			if (!findServer(network.allServers, dhcpData.ip))
			{
				hex2String(mapname, dhcpData.bp_chaddr, dhcpData.bp_hlen);
				data7 *dhcpEntry = findDHCPEntry(mapname);

				if (dhcpEntry)
				{
					if (dhcpEntry->ip != dhcpData.ip)
						continue;
				}
				else
				{
					dhcpMap::iterator p = dhcpCache.begin();

					for (; p != dhcpCache.end(); p++)
					{
						if (p->second && p->second->ip == dhcpData.ip && strcasecmp(mapname, p->second->mapname))
							break;
					}

					if (p != dhcpCache.end())
						continue;
				}

				rangeInd = getRangeInd(dhcpData.ip);

				if(rangeInd >= 0)
				{
					ind = getIndex(rangeInd, dhcpData.ip);

					if (ind >= 0 && !dhcpEntry)
					{
						dhcpEntry = (data7*)calloc(1, sizeof(data7));

						if (!dhcpEntry)
						{
							sprintf(logBuff, "Loading Existing Leases, Memory Allocation Error");
							logDHCPMess(logBuff, 1);
							return;
						}

						dhcpEntry->mapname = cloneString(mapname);

						if (!dhcpEntry->mapname)
						{
							sprintf(logBuff, "Loading Existing Leases, Memory Allocation Error");
							free(dhcpEntry);
							logDHCPMess(logBuff, 1);
							return;
						}

						dhcpCache[dhcpEntry->mapname] = dhcpEntry;
					}
				}

				if (dhcpEntry)
				{
					dhcpEntry->ip = dhcpData.ip;
					dhcpEntry->rangeInd = rangeInd;
					dhcpEntry->expiry = dhcpData.expiry;
					dhcpEntry->hostname = cloneString(dhcpData.hostname);
					hangDhcpEntry(dhcpEntry);
				}
			}
		}

		fclose(f);

		f = fopen(leaFile, "wb");
		cfig.dhcpInd = 0;

		if (f)
		{
			dhcpMap::iterator p = dhcpCache.begin();

			for (; p != dhcpCache.end(); p++)
			{
				if ((dhcpEntry = p->second) && (dhcpEntry->expiry > t || !dhcpEntry->fixed))
				{
					memset(&dhcpData, 0, sizeof(data8));
					dhcpData.bp_hlen = 16;
					getHexValue(dhcpData.bp_chaddr, dhcpEntry->mapname, &dhcpData.bp_hlen);
					dhcpData.ip = dhcpEntry->ip;
					dhcpData.expiry = dhcpEntry->expiry;
					strcpy(dhcpData.hostname, dhcpEntry->hostname);
					cfig.dhcpInd++;
					dhcpData.dhcpInd = cfig.dhcpInd;
					dhcpEntry->dhcpInd = cfig.dhcpInd;
					fwrite(&dhcpData, sizeof(data8), 1, f);
				}
			}
			fclose(f);
		}
	}
}

bool getSection(const char *sectionName, char *buffer, MYBYTE index, char *fileName)
{
	//printf("%s=%s\n",fileName,sectionName);
	char section[128];
	sprintf(section, "[%s]", sectionName);
	myUpper(section);
	FILE *f = fopen(fileName, "rt");
	char buff[512];
	MYBYTE found = 0;

	if (f)
	{
		while (fgets(buff, 511, f))
		{
			myUpper(buff);
			myTrim(buff, buff);

			if (strstr(buff, section) == buff)
			{
				found++;
				if (found == index)
				{
					//printf("%s=%s\n",fileName,sectionName);
					while (fgets(buff, 511, f))
					{
						myTrim(buff, buff);

						if (strstr(buff, "[") == buff)
							break;

						if ((*buff) >= '0' && (*buff) <= '9' || (*buff) >= 'A' && (*buff) <= 'Z' || (*buff) >= 'a' && (*buff) <= 'z' || ((*buff) && strchr("/\\?*", (*buff))))
						{
							buffer += sprintf(buffer, "%s", buff);
							buffer++;
						}
					}
					break;
				}
			}
		}
		fclose(f);
	}

	*buffer = 0;
	*(buffer + 1) = 0;
	return (found == index);
}

FILE *openSection(const char *sectionName, MYBYTE index, char *fileName)
{
	//printf("%s=%s\n",fileName,sectionName);
	char section[128];
	sprintf(section, "[%s]", sectionName);
	myUpper(section);
	FILE *f = fopen(fileName, "rt");
	char raw[512];
	MYBYTE found = 0;

	if (f)
	{
		while (fgets(raw, 511, f))
		{
			myUpper(raw);
			myTrim(raw, raw);

			if (strstr(raw, section) == raw)
			{
				found++;

				if (found == index)
					return f;
			}
		}
		fclose(f);
	}
	return NULL;
}

char *readSection(char* raw, FILE *f)
{
	while (fgets(raw, 511, f))
	{
		myTrim(raw, raw);

		if (*raw == '[')
			break;

		if ((*raw) >= '0' && (*raw) <= '9' || (*raw) >= 'A' && (*raw) <= 'Z' || (*raw) >= 'a' && (*raw) <= 'z' || ((*raw) && strchr("/\\?*", (*raw))))
			return raw;
	}

	fclose(f);
	return NULL;
}

char* myGetToken(char* buff, MYBYTE index)
{
	while (*buff)
	{
		if (index)
			index--;
		else
			break;

		buff += strlen(buff) + 1;
	}

	return buff;
}

MYWORD myTokenize(char *target, char *source, const char *sep, bool whiteSep)
{
	bool found = true;
	char *dp = target;
	MYWORD kount = 0;

	while (*source)
	{
		if (sep && sep[0] && strchr(sep, (*source)))
		{
			found = true;
			source++;
			continue;
		}
		else if (whiteSep && (*source) <= 32)
		{
			found = true;
			source++;
			continue;
		}

		if (found)
		{
			if (target != dp)
			{
				*dp = 0;
				dp++;
			}
			kount++;
		}

		found = false;
		*dp = *source;
		dp++;
		source++;
	}

	*dp = 0;
	dp++;
	*dp = 0;

	//printf("%s\n", target);

	return kount;
}

char* myTrim(char *target, char *source)
{
	while ((*source) && (*source) <= 32)
		source++;

	int i = 0;

	for (; i < 511 && source[i]; i++)
		target[i] = source[i];

	target[i] = source[i];
	i--;

	for (; i >= 0 && target[i] <= 32; i--)
		target[i] = 0;

	return target;
}

void mySplit(char *name, char *value, char *source, char splitChar)
{
	int i = 0;
	int j = 0;
	int k = 0;

	for (; source[i] && j <= 510 && source[i] != splitChar; i++, j++)
	{
		name[j] = source[i];
	}

	if (source[i])
	{
		i++;
		for (; k <= 510 && source[i]; i++, k++)
		{
			value[k] = source[i];
		}
	}

	name[j] = 0;
	value[k] = 0;

	myTrim(name, name);
	myTrim(value, value);
	//printf("%s %s\n", name, value);
}


char *strqtype(char *buff, MYBYTE qtype)
{
	switch (qtype)
	{
		case 1:
			strcpy(buff, "A");
			break;
		case 2:
			strcpy(buff, "NS");
			break;
		case 3:
			strcpy(buff, "MD");
			break;
		case 4:
			strcpy(buff, "MF");
			break;
		case 5:
			strcpy(buff, "CNAME");
			break;
		case 6:
			strcpy(buff, "SOA");
			break;
		case 7:
			strcpy(buff, "MB");
			break;
		case 8:
			strcpy(buff, "MG");
			break;
		case 9:
			strcpy(buff, "MR");
			break;
		case 10:
			strcpy(buff, "NULL");
			break;
		case 11:
			strcpy(buff, "WKS");
			break;
		case 12:
			strcpy(buff, "PTR");
			break;
		case 13:
			strcpy(buff, "HINFO");
			break;
		case 14:
			strcpy(buff, "MINFO");
			break;
		case 15:
			strcpy(buff, "MX");
			break;
		case 16:
			strcpy(buff, "TXT");
			break;
		case 28:
			strcpy(buff, "AAAA");
			break;
		case 251:
			strcpy(buff, "IXFR");
			break;
		case 252:
			strcpy(buff, "AXFR");
			break;
		case 253:
			strcpy(buff, "MAILB");
			break;
		case 254:
			strcpy(buff, "MAILA");
			break;
		default:
			strcpy(buff, "ANY");
			break;
	}
	return buff;
}

MYDWORD getClassNetwork(MYDWORD ip)
{
	data15 data;
	data.ip = ip;
	data.octate[3] = 0;

	if (data.octate[0] < 192)
		data.octate[2] = 0;

	if (data.octate[0] < 128)
		data.octate[1] = 0;

	return data.ip;
}

/*
char *IP2Auth(MYDWORD ip)
{
data15 data;
data.ip = ip;

if (data.octate[0] >= 192)
sprintf(tempbuff, "%u.%u.%u", data.octate[2], data.octate[1], data.octate[0]);
else if (data.octate[0] >= 128)
sprintf(tempbuff, "%u.%u", data.octate[1], data.octate[0]);
else
sprintf(tempbuff, "%u", data.octate[0]);

strcat(tempbuff, arpa);
return tempbuff;
}
*/

char *IP2String(char *target, MYDWORD ip)
{
	data15 inaddr;
	inaddr.ip = ip;
	sprintf(target, "%u.%u.%u.%u", inaddr.octate[0], inaddr.octate[1], inaddr.octate[2], inaddr.octate[3]);
	//MYBYTE *octate = (MYBYTE*)&ip;
	//sprintf(target, "%u.%u.%u.%u", octate[0], octate[1], octate[2], octate[3]);
	return target;
}

bool addServer(MYDWORD *array, MYDWORD ip)
{
	for (MYBYTE i = 0; i < MAX_SERVERS; i++)
	{
		if (!ip || array[i] == ip)
			return false;
		else if (!array[i])
		{
			array[i] = ip;
			return true;
		}
	}
	return false;
}

MYDWORD *findServer(MYDWORD *array, MYDWORD ip)
{
	if (ip)
	{
		for (MYBYTE i = 0; i < MAX_SERVERS && array[i]; i++)
		{
			if (array[i] == ip)
				return &(array[i]);
		}
	}
	return 0;
}

bool isInt(char *str)
{
	if (!str || !(*str))
		return false;

	for(; *str; str++)
		if (*str <  '0' || *str > '9')
			return false;

	return true;
}

bool isIP(char *str)
{
	if (!str || !(*str))
		return false;

	MYDWORD ip = inet_addr(str);
	int j = 0;

	for (; *str; str++)
	{
		if (*str == '.' && *(str + 1) != '.')
			j++;
		else if (*str < '0' || *str > '9')
			return false;
	}

	if (j == 3)
	 {
		if (ip == INADDR_NONE || ip == INADDR_ANY)
			return false;
		else
			return true;
	}
	else
		return false;
}

/*
char *toBase64(MYBYTE *source, MYBYTE length)
{
	MYBYTE a = 0, b = 0, i = 0;
	char *dp = tempbuff;

	for (; length; length--, source++)
	{
		i += 2;
		a = (*source) >> i;
		*dp = base64[a + b];
		dp++;
		b = (*source) << (8 - i);
		b >>= 2;
		if (i == 6)
		{
			*dp = base64[b];
			dp++;
			i = b = 0;
		}
	}
	if (i)
	{
		*dp = base64[b];
		dp++;
	}
	*dp = 0;
	//printf("%s\n",tempbuff);
	return tempbuff;
}

MYBYTE getBaseValue(MYBYTE a)
{
	if (a >= 'A' && a <= 'Z')
		a -= 'A';
	else if (a >= 'a' && a <= 'z')
		a = a - 'a' + 26;
	else if (a >= '0' && a <= '9')
		a = a - '0' + 52;
	else if (a == '+')
		a = 62;
	else if (a == '/')
		a = 63;
	else
		a = UCHAR_MAX;

	return a;
}

MYBYTE fromBase64(MYBYTE *target, char *source)
{
	//printf("SOURCE=%s\n", source);
	MYBYTE b = 0;
	MYBYTE shift = 4;
	MYBYTE bp_hlen = (3 * strlen(source))/4;
	*target = 0;

	if (*source)
	{
		b = getBaseValue(*source);
		*target = b << 2;
		source++;

		while (*source)
		{
			b = getBaseValue(*source);
			(*target) += (b >> (8 - shift));
			target++;
			(*target) = (b << shift);
			shift += 2;

			if (shift > 8)
			{
				source++;

				if (*source)
				{
					b = getBaseValue(*source);
					*target = b << 2;
					shift = 4;
				}
				else
					break;
			}

			source++;
		}
	}
	//printf("SIZE=%u\n", bp_hlen);
	return bp_hlen;
}

char *toUUE(char *tempbuff, MYBYTE *source, MYBYTE length)
{
	MYBYTE a = 0, b = 0, i = 0;
	char *dp = tempbuff;

	for (; length; length--, source++)
	{
		i += 2;
		a = (*source) >> i;
		*dp = a + b + 32;
		dp++;
		b = (*source) << (8 - i);
		b >>= 2;
		if (i == 6)
		{
			*dp = b + 32;
			dp++;
			i = b = 0;
		}
	}
	if (i)
	{
		*dp = b + 32;
		dp++;
	}
	*dp = 0;
	//printf("%s\n",tempbuff);
	return tempbuff;
}

MYBYTE fromUUE(MYBYTE *target, char *source)
{
	//printf("SOURCE=%s\n", source);
	MYBYTE b = 0;
	MYBYTE shift = 4;
	MYBYTE bp_hlen = (3 * strlen(source))/4;
	*target = 0;

	if (*source)
	{
		b = *source - 32;
		*target = b << 2;
		source++;

		while (*source)
		{
			b = *source - 32;
			(*target) += (b >> (8 - shift));
			target++;
			(*target) = (b << shift);
			shift += 2;

			if (shift > 8)
			{
				source++;

				if (*source)
				{
					b = *source - 32;
					*target = b << 2;
					shift = 4;
				}
				else
					break;
			}

			source++;
		}
	}
	//printf("SIZE=%u\n", bp_hlen);
	return bp_hlen;
}
*/

char *hex2String(char *target, MYBYTE *hex, MYBYTE bytes)
{
	char *dp = target;

	if (bytes)
		dp += sprintf(target, "%02x", *hex);
	else
		*target = 0;

	for (MYBYTE i = 1; i < bytes; i++)
			dp += sprintf(dp, ":%02x", *(hex + i));

	return target;
}

char *genHostName(char *target, MYBYTE *hex, MYBYTE bytes)
{
	char *dp = target;

	if (bytes)
		dp += sprintf(target, "Host%02x", *hex);
	else
		*target = 0;

	for (MYBYTE i = 1; i < bytes; i++)
			dp += sprintf(dp, "%02x", *(hex + i));

	return target;
}

/*
char *IP62String(char *target, MYBYTE *source)
{
	MYWORD *dw = (MYWORD*)source;
	char *dp = target;
	MYBYTE markbyte;

	for (markbyte = 4; markbyte > 0 && !dw[markbyte - 1]; markbyte--);

	for (MYBYTE i = 0; i < markbyte; i++)
		dp += sprintf(dp, "%x:", ntohs(dw[i]));

	for (markbyte = 4; markbyte < 8 && !dw[markbyte]; markbyte++);

	for (MYBYTE i = markbyte; i < 8; i++)
		dp += sprintf(dp, ":%x", htons(dw[i]));

	return target;
}
*/

char *IP62String(char *target, MYBYTE *source)
{
	char *dp = target;
	bool zerostarted = false;
	bool zeroended = false;

	for (MYBYTE i = 0; i < 16; i += 2, source += 2)
	{
		if (source[0])
		{
			if (zerostarted)
				zeroended = true;

			if (zerostarted && zeroended)
			{
				dp += sprintf(dp, "::");
				zerostarted = false;
			}
			else if (dp != target)
				dp += sprintf(dp, ":");

			dp += sprintf(dp, "%x", source[0]);
			dp += sprintf(dp, "%02x", source[1]);
		}
		else if (source[1])
		{
			if (zerostarted)
				zeroended = true;

			if (zerostarted && zeroended)
			{
				dp += sprintf(dp, "::");
				zerostarted = false;
			}
			else if (dp != target)
				dp += sprintf(dp, ":");

			dp += sprintf(dp, "%0x", source[1]);
		}
		else if (!zeroended)
			zerostarted = true;
	}

	return target;
}

char *getHexValue(MYBYTE *target, char *source, MYBYTE *size)
{
	if (*size)
		memset(target, 0, (*size));

	if (*source == ':')
		source++;

	for ((*size) = 0; (*source) && (*size) < UCHAR_MAX; (*size)++, target++)
	{
		if ((*source) >= '0' && (*source) <= '9')
		{
			(*target) = (*source) - '0';
		}
		else if ((*source) >= 'a' && (*source) <= 'f')
		{
			(*target) = (*source) - 'a' + 10;
		}
		else if ((*source) >= 'A' && (*source) <= 'F')
		{
			(*target) = (*source) - 'A' + 10;
		}
		else
		{
			return source;
		}

		source++;

		if ((*source) >= '0' && (*source) <= '9')
		{
			(*target) *= 16;
			(*target) += ((*source) - '0');
		}
		else if ((*source) >= 'a' && (*source) <= 'f')
		{
			(*target) *= 16;
			(*target) += (((*source) - 'a') + 10);
		}
		else if ((*source) >= 'A' && (*source) <= 'F')
		{
			(*target) *= 16;
			(*target) += (((*source) - 'A') + 10);
		}
		else if ((*source) == ':')
		{
			source++;
			continue;
		}
		else if (*source)
			return source;
		else
			continue;

		source++;

		if ((*source) == ':')
			source++;
		else if (*source)
			return source;
	}

	if (*source)
		return source;

	return NULL;
}

char *myUpper(char *string)
{
	char diff = 'a' - 'A';
	MYWORD len = strlen(string);
	for (int i = 0; i < len; i++)
		if (string[i] >= 'a' && string[i] <= 'z')
			string[i] -= diff;
	return string;
}

char *myLower(char *string)
{
	char diff = 'a' - 'A';
	MYWORD len = strlen(string);
	for (int i = 0; i < len; i++)
		if (string[i] >= 'A' && string[i] <= 'Z')
			string[i] += diff;
	return string;
}

bool wildcmp(char *string, char *wild)
{
	// Written by Jack Handy - jakkhandy@hotmail.com
	// slightly modified
	char *cp = NULL;
	char *mp = NULL;

	while ((*string) && (*wild != '*'))
	{
		if ((*wild != *string) && (*wild != '?'))
		{
			return 0;
		}
		wild++;
		string++;
	}

	while (*string)
	{
		if (*wild == '*')
		{
			if (!*++wild)
				return 1;

			mp = wild;
			cp = string + 1;
		}
		else if ((*wild == *string) || (*wild == '?'))
		{
			wild++;
			string++;
		}
		else
		{
			wild = mp;
			string = cp++;
		}
	}

	while (*wild == '*')
		wild++;

	return !(*wild);
}

/*
void checkSize()
{
	data7 *dhcpEntry = NULL;

	//printf("checksize dhcpdhcpEntry[%d]=%d dhcpAge[%d]=%d\n", currentInd, dhcpdhcpEntry.size(), currentInd, dhcpAge.size());
	expiryMap::iterator p = dhcpAge.begin();
	expiryMap::iterator q;

	while (p != dhcpAge.end() && p->first < t)
	{
		dhcpEntry = p->second;
		//printf("processing %s=%d\n", dhcpEntry->mapname, p->first - t);

		if (dhcpEntry->expiry < t)
		{
			sprintf(logBuff, "Lease %s Being Released", IP2String(tempbuff, dhcpEntry->ip));
			logDHCPMess(logBuff, 2);
			q = p;
			p++;

			dhcpAge.erase(q);
			dhcpEntry->hanged = false;
			updateDHCP(dhcpEntry, 0);
		}
		else if (dhcpEntry->expiry > p->first)
		{
			q = p;
			p++;
			dhcpAge.erase(q);
			dhcpAge.insert(pair<time_t, data7*>(dhcpEntry->expiry, dhcpEntry));
		}
		else
			p++;
	}
	//printf("Done %u=%u\n",dnsdhcpEntry[ind].size(),dnsAge[currentInd].size());
}
*/

void calcRangeLimits(MYDWORD ip, MYDWORD mask, MYDWORD *rangeStart, MYDWORD *rangeEnd)
{
	*rangeStart = htonl(ip & mask) + 1;
	*rangeEnd = htonl(ip | (~mask)) - 1;
}

bool checkMask(MYDWORD mask)
{
	mask = htonl(mask);
	while (mask)
	{
		if (mask < (mask << 1))
			return false;

		mask <<= 1;
	}
	return true;
}

MYDWORD calcMask(MYDWORD rangeStart, MYDWORD rangeEnd)
{
	data15 ip1, ip2, mask;

	ip1.ip = htonl(rangeStart);
	ip2.ip = htonl(rangeEnd);

	for (MYBYTE i = 0; i < 4; i++)
	{
		mask.octate[i] = ip1.octate[i] ^ ip2.octate[i];

		if (i && mask.octate[i - 1] < 255)
			mask.octate[i] = 0;
		else if (mask.octate[i] == 0)
			mask.octate[i] = 255;
		else if (mask.octate[i] < 2)
			mask.octate[i] = 254;
		else if (mask.octate[i] < 4)
			mask.octate[i] = 252;
		else if (mask.octate[i] < 8)
			mask.octate[i] = 248;
		else if (mask.octate[i] < 16)
			mask.octate[i] = 240;
		else if (mask.octate[i] < 32)
			mask.octate[i] = 224;
		else if (mask.octate[i] < 64)
			mask.octate[i] = 192;
		else if (mask.octate[i] < 128)
			mask.octate[i] = 128;
		else
			mask.octate[i] = 0;
	}
	return mask.ip;
}

char *cloneString(char *string)
{
	if (!string)
		return NULL;

	char *s = (char*)calloc(1, strlen(string) + 1);

	if (s)
	{
		strcpy(s, string);
	}
	return s;
}

void *init(void *lparam)
{
    char raw[512];
    char name[512];
    char value[512];
    FILE *f = NULL;

	if (verbatim)
	{
		cfig.dhcpLogLevel = 2;
		printf("%s\n\n", sVersion);
	}
	else if (f = openSection("LOGGING", 1, iniFile))
	{
		cfig.dhcpLogLevel = 1;
		tempbuff[0] = 0;

		while (readSection(raw, f))
		{
			mySplit(name, value, raw, '=');

			if (name[0] && value[0])
			{
				if (!strcasecmp(name, "LogLevel"))
				{
					if (!strcasecmp(value, "None"))
						cfig.dhcpLogLevel = 0;
					else if (!strcasecmp(value, "Normal"))
						cfig.dhcpLogLevel = 1;
					else if (!strcasecmp(value, "All"))
						cfig.dhcpLogLevel = 2;
					else
						sprintf(tempbuff, "Section [LOGGING], Invalid LogLevel: %s", value);
				}
				else
					sprintf(tempbuff, "Section [LOGGING], Invalid Entry %s ignored", raw);
			}
			else
				sprintf(tempbuff, "Section [LOGGING], Invalid Entry %s ignored", raw);
		}
	}

	if ((f = fopen(iniFile, "rt")))
	{
		fclose(f);
	}
	else
	{
		sprintf(logBuff, "Warning: file %s not found, defaults will be used", iniFile);
		logDHCPMess(logBuff, 1);
	}

	if (!verbatim && cfig.dhcpLogLevel && logFile[0])
	{
		tm *ttm = localtime(&t);
		loggingDay = ttm->tm_yday;
		strftime(extbuff, sizeof(extbuff), logFile, ttm);

		cfig.logfile = fopen(extbuff, "at");

		if (cfig.logfile)
		{
			sprintf(logBuff, "%s Starting..", sVersion);
			logDHCPMess(logBuff, 1);

			if (tempbuff[0])
				logDHCPMess(tempbuff, 0);
		}
    }

	loadDHCP();

	if (verbatim)
		printf("Starting DHCP...\n");
	else
	{
		sprintf(logBuff, "Starting DHCP Service");
		logDHCPMess(logBuff, 1);

		if (cfig.dhcpLogLevel > 1)
			sprintf(logBuff, "Logging: All");
		else if (cfig.dhcpLogLevel >= 1)
			sprintf(logBuff, "Logging: Normal");
		else
			sprintf(logBuff, "Logging: None");

		logDHCPMess(logBuff, 1);
	}

	for (int i = 0; i < MAX_RANGES && cfig.dhcpRanges[i].rangeStart;i++)
	{
		char *logPtr = logBuff;
		logPtr += sprintf(logPtr, "DHCP Range: ");
		logPtr += sprintf(logPtr, "%s", IP2String(tempbuff, htonl(cfig.dhcpRanges[i].rangeStart)));
		logPtr += sprintf(logPtr, "-%s", IP2String(tempbuff, htonl(cfig.dhcpRanges[i].rangeEnd)));
		logPtr += sprintf(logPtr, "/%s", IP2String(tempbuff, cfig.dhcpRanges[i].mask));
		logDHCPMess(logBuff, 1);
	}

    gethostname(cfig.servername, sizeof(cfig.servername));

	char *ptr = strchr(cfig.servername, '.');

	if (ptr)
	{
		*ptr = 0;
		ptr++;
		strcpy(cfig.zone, ptr);
		cfig.zLen = strlen(cfig.zone);
	}
#ifdef getdomainname
	else
	{
		getdomainname(cfig.zone, sizeof(cfig.zone));
		cfig.zLen = strlen(cfig.zone);
	}
#endif

	cfig.fixedSocket = socket(AF_INET, SOCK_DGRAM, 0);

	if (cfig.fixedSocket < 0)
	{
		sprintf(logBuff, "Failed to Create Socket");
		logDHCPMess(logBuff, 1);
		exit(EXIT_FAILURE);
	}

	getInterfaces(&newNetwork);
	memcpy(cfig.oldservers, newNetwork.staticServers, (MAX_SERVERS * sizeof(MYDWORD)));

	if (f = openSection("REPLICATION_SERVERS", 1, iniFile))
	{
		while (readSection(raw, f))
		{
			mySplit(name, value, raw, '=');

			if (name[0] && value[0])
			{
				if (!isIP(name) && isIP(value))
				{
					if (!strcasecmp(name, "Primary"))
						cfig.zoneServers[0] = inet_addr(value);
					else if (!strcasecmp(name, "Secondary"))
						cfig.zoneServers[1] = inet_addr(value);
					else
					{
						sprintf(logBuff, "Section [REPLICATION_SERVERS] Invalid Entry: %s ignored", raw);
						logDHCPMess(logBuff, 1);
					}
				}
				else
				{
					sprintf(logBuff, "Section [REPLICATION_SERVERS] Invalid Entry: %s ignored", raw);
					logDHCPMess(logBuff, 1);
				}
			}
			else
			{
				sprintf(logBuff, "Section [REPLICATION_SERVERS], Missing value, entry %s ignored", raw);
				logDHCPMess(logBuff, 1);
			}
		}
	}

	if (!cfig.zoneServers[0] && cfig.zoneServers[1])
	{
		sprintf(logBuff, "Section [REPLICATION_SERVERS] Missing Primary Server");
		logDHCPMess(logBuff, 1);
	}
	else if (cfig.zoneServers[0] && !cfig.zoneServers[1])
	{
		sprintf(logBuff, "Section [REPLICATION_SERVERS] Missing Secondary Server");
		logDHCPMess(logBuff, 1);
	}
	else if (cfig.zoneServers[0] && cfig.zoneServers[1])
	{
		if (findServer(newNetwork.staticServers, cfig.zoneServers[0]) && findServer(newNetwork.staticServers, cfig.zoneServers[1]))
		{
			sprintf(logBuff, "Section [REPLICATION_SERVERS] Primary & Secondary should be Different Boxes");
			logDHCPMess(logBuff, 1);
		}
		else if (findServer(newNetwork.staticServers, cfig.zoneServers[0]))
			cfig.replication = 1;
		else if (findServer(newNetwork.staticServers, cfig.zoneServers[1]))
			cfig.replication = 2;
		else
		{
			sprintf(logBuff, "Section [REPLICATION_SERVERS] No Server IP not found on this Machine");
			logDHCPMess(logBuff, 1);
		}
	}

	if (cfig.replication)
	{
		//sprintf(logBuff, "Contacting Primary DHCP Server %s", IP2String(tempbuff, cfig.zoneServers[0]));
		//logDHCPMess(logBuff, 1);

		cfig.dhcpReplConn.sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

		if (cfig.dhcpReplConn.sock == INVALID_SOCKET)
		{
			sprintf(logBuff, "Failed to Create DHCP Replication Socket");
			logDHCPMess(logBuff, 1);
		}
		else
		{
			//printf("Socket %u\n", cfig.dhcpReplConn.sock);

			if (cfig.replication == 1)
				cfig.dhcpReplConn.server = cfig.zoneServers[0];
			else
				cfig.dhcpReplConn.server = cfig.zoneServers[1];

			cfig.dhcpReplConn.addr.sin_family = AF_INET;
			cfig.dhcpReplConn.addr.sin_addr.s_addr = cfig.dhcpReplConn.server;
			cfig.dhcpReplConn.addr.sin_port = 0;

			int nRet = bind(cfig.dhcpReplConn.sock, (sockaddr*)&cfig.dhcpReplConn.addr, sizeof(struct sockaddr_in));

			if (nRet == SOCKET_ERROR)
			{
				cfig.dhcpReplConn.ready = false;
				sprintf(logBuff, "DHCP Replication Server, Bind Failed");
				logDHCPMess(logBuff, 1);
			}
			else
			{
				cfig.dhcpReplConn.port = IPPORT_DHCPS;
				cfig.dhcpReplConn.loaded = true;
				cfig.dhcpReplConn.ready = true;

				data3 op;
				memset(&token, 0, sizeof(data9));
				token.vp = token.dhcpp.vend_data;
				token.messsize = sizeof(dhcp_packet);

				token.dhcpp.header.bp_op = BOOTP_REQUEST;
				token.dhcpp.header.bp_xid = t;
				token.dhcpp.header.bp_magic_num[0] = 99;
				token.dhcpp.header.bp_magic_num[1] = 130;
				token.dhcpp.header.bp_magic_num[2] = 83;
				token.dhcpp.header.bp_magic_num[3] = 99;

				op.opt_code = DHCP_OPTION_MESSAGETYPE;
				op.size = 1;
				op.value[0] = DHCP_MESS_INFORM;
				pvdata(&token, &op);

				token.vp[0] = DHCP_OPTION_END;
				token.vp++;

				token.remote.sin_port = htons(IPPORT_DHCPS);
				token.remote.sin_family = AF_INET;

				if (cfig.replication == 1)
					token.remote.sin_addr.s_addr = cfig.zoneServers[1];
				else
					token.remote.sin_addr.s_addr = cfig.zoneServers[0];

				if (cfig.replication == 2)
				{
					pthread_t threadId;
					pthread_attr_t attr;
					pthread_attr_init(&attr);
					pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
					int errcode = pthread_create(&threadId, &attr, sendToken, 0);
					pthread_attr_destroy(&attr);
				}
			}
		}
	}

	sprintf(cfig.servername_fqn, "%s.%s", cfig.servername, cfig.zone);

	if (cfig.replication == 1)
		sprintf(logBuff, "Server Name: %s (Primary)", cfig.servername);
	else if (cfig.replication == 2)
		sprintf(logBuff, "Server Name: %s (Secondary)", cfig.servername);
	else
		sprintf(logBuff, "Server Name: %s", cfig.servername);

	logDHCPMess(logBuff, 1);

	sprintf(logBuff, "Detecting Static Interfaces..");
	logDHCPMess(logBuff, 1);

	do
	{
		if (newNetwork.maxFD < cfig.dhcpReplConn.sock)
			newNetwork.maxFD = cfig.dhcpReplConn.sock;

		bool bindfailed = false;

		if (f = openSection("LISTEN_ON", 1, iniFile))
		{
			//char *name = myGetToken(iniStr, 0)
			int i = 0;

			while (readSection(raw, f))
			{
				if(i < MAX_SERVERS)
				{
					cfig.ifspecified = true;
					MYDWORD addr = inet_addr(raw);

					if (isIP(raw) && addr)
					{
						for (MYBYTE m = 0; ; m++)
						{
							if (m >= MAX_SERVERS || !newNetwork.staticServers[m])
							{
								if (findServer(newNetwork.allServers, addr))
								{
									sprintf(logBuff, "Warning: Section [LISTEN_ON], Interface %s is not Static, ignored", raw);
									logDHCPMess(logBuff, 1);
								}
								else
								{
									bindfailed = true;
									sprintf(logBuff, "Warning: Section [LISTEN_ON], Interface %s not available, ignored", raw);
									logDHCPMess(logBuff, 1);
								}
								break;
							}
							else if (newNetwork.staticServers[m] == addr)
							{
								for (MYBYTE n = 0; n < MAX_SERVERS; n++)
								{
									if (newNetwork.listenServers[n] == addr)
										break;
									else if (!newNetwork.listenServers[n])
									{
										newNetwork.listenServers[n] = newNetwork.staticServers[m];
										newNetwork.listenMasks[n] = newNetwork.staticMasks[m];
										break;
									}
								}
								break;
							}
						}
					}
					else
					{
						sprintf(logBuff, "Warning: Section [LISTEN_ON], Invalid Interface Address %s, ignored", raw);
						logDHCPMess(logBuff, 1);
					}
				}
			}
		}

		if (!cfig.ifspecified)
		{
			for (MYBYTE n = 0; n < MAX_SERVERS && newNetwork.staticServers[n]; n++)
			{
				newNetwork.listenServers[n] = newNetwork.staticServers[n];
				newNetwork.listenMasks[n] = newNetwork.staticMasks[n];
			}
		}

		int i = 0;

		for (int j = 0; j < MAX_SERVERS && newNetwork.listenServers[j]; j++)
		{
			int k = 0;

			for (; k < MAX_SERVERS && network.dhcpConn[k].loaded; k++)
			{
				if (network.dhcpConn[k].ready && network.dhcpConn[k].server == newNetwork.listenServers[j])
					break;
			}

			if (network.dhcpConn[k].ready && network.dhcpConn[k].server == newNetwork.listenServers[j])
			{
				memcpy(&(newNetwork.dhcpConn[i]), &(network.dhcpConn[k]), sizeof(DhcpConnType));

				if (newNetwork.maxFD < newNetwork.dhcpConn[i].sock)
					newNetwork.maxFD = newNetwork.dhcpConn[i].sock;

				network.dhcpConn[k].ready = false;
				i++;
				continue;
			}
			else
			{
				newNetwork.dhcpConn[i].sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

				if (newNetwork.dhcpConn[i].sock == INVALID_SOCKET)
				{
					bindfailed = true;
					sprintf(logBuff, "Failed to Create Socket");
					logDHCPMess(logBuff, 1);
					continue;
				}

				//printf("Socket %u\n", newNetwork.dhcpConn[i].sock);

				newNetwork.dhcpConn[i].addr.sin_family = AF_INET;
				newNetwork.dhcpConn[i].addr.sin_addr.s_addr = newNetwork.listenServers[j];
				newNetwork.dhcpConn[i].addr.sin_port = htons(IPPORT_DHCPS);
				newNetwork.dhcpConn[i].reUseVal = true;
				newNetwork.dhcpConn[i].reUseSize = sizeof(newNetwork.dhcpConn[i].reUseVal);

				setsockopt(newNetwork.dhcpConn[i].sock, SOL_SOCKET, SO_REUSEADDR, (char*)&newNetwork.dhcpConn[i].reUseVal, newNetwork.dhcpConn[i].reUseSize);

				newNetwork.dhcpConn[i].broadCastVal = true;
				newNetwork.dhcpConn[i].broadCastSize = sizeof(newNetwork.dhcpConn[i].broadCastVal);

				setsockopt(newNetwork.dhcpConn[i].sock, SOL_SOCKET, SO_BROADCAST, (char*)&newNetwork.dhcpConn[i].broadCastVal, newNetwork.dhcpConn[i].broadCastSize);
				int nRet = bind(newNetwork.dhcpConn[i].sock, (sockaddr*)&newNetwork.dhcpConn[i].addr, sizeof(struct sockaddr_in));

				if (nRet == SOCKET_ERROR)
				{
					close(newNetwork.dhcpConn[i].sock);
					sprintf(logBuff, "%s Port 67 already in use", IP2String(tempbuff, newNetwork.listenServers[j]));
					logDHCPMess(logBuff, 1);
					continue;
				}

				newNetwork.dhcpConn[i].loaded = true;
				newNetwork.dhcpConn[i].ready = true;

				if (newNetwork.maxFD < newNetwork.dhcpConn[i].sock)
					newNetwork.maxFD = newNetwork.dhcpConn[i].sock;

				newNetwork.dhcpConn[i].server = newNetwork.listenServers[j];
				newNetwork.dhcpConn[i].mask = newNetwork.listenMasks[j];
				newNetwork.dhcpConn[i].port = IPPORT_DHCPS;

				i++;
			}
		}

		if (network.dhcpListener.ready)
		{
			memcpy(&(newNetwork.dhcpListener), &(network.dhcpListener), sizeof(DhcpConnType));

			if (newNetwork.maxFD < newNetwork.dhcpListener.sock)
				newNetwork.maxFD = newNetwork.dhcpListener.sock;

			network.dhcpListener.ready = false;
		}
		else
		{
			newNetwork.dhcpListener.sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

			if (newNetwork.dhcpListener.sock == INVALID_SOCKET)
			{
				sprintf(logBuff, "Failed to Creat DHCP Socket");
				logDHCPMess(logBuff, 1);
			}
			else
			{
				newNetwork.dhcpListener.reUseVal = true;
				newNetwork.dhcpListener.reUseSize = sizeof(newNetwork.dhcpListener.reUseVal);
				setsockopt(newNetwork.dhcpListener.sock, SOL_SOCKET, SO_REUSEADDR, (char*)&newNetwork.dhcpListener.reUseVal, newNetwork.dhcpListener.reUseSize);
				newNetwork.dhcpListener.pktinfoVal = true;
				newNetwork.dhcpListener.pktinfoSize = sizeof(newNetwork.dhcpListener.pktinfoVal);
				setsockopt(newNetwork.dhcpListener.sock, SOL_IP, IP_PKTINFO, &newNetwork.dhcpListener.pktinfoVal, newNetwork.dhcpListener.pktinfoSize);

				newNetwork.dhcpListener.addr.sin_family = AF_INET;
				newNetwork.dhcpListener.addr.sin_addr.s_addr = INADDR_ANY;
				newNetwork.dhcpListener.addr.sin_port = htons(IPPORT_DHCPS);

				int nRet = bind(newNetwork.dhcpListener.sock,
								(sockaddr*)&newNetwork.dhcpListener.addr,
								sizeof(struct sockaddr_in)
							   );

				if (nRet == SOCKET_ERROR)
				{
					close(newNetwork.dhcpListener.sock);
					sprintf(logBuff, "DHCP Port 67 already in use");
					logDHCPMess(logBuff, 1);
				}
				else
				{
					newNetwork.dhcpListener.loaded = true;
					newNetwork.dhcpListener.ready = true;

					if (newNetwork.maxFD < newNetwork.dhcpListener.sock)
						newNetwork.maxFD = newNetwork.dhcpListener.sock;
				}
			}
		}

		newNetwork.httpConn.port = 6789;
		newNetwork.httpConn.server = newNetwork.dhcpConn[0].server;
		newNetwork.httpConn.loaded = true;

		if (f = openSection("HTTP_INTERFACE", 1, iniFile))
		{
			while (readSection(raw, f))
			{
				mySplit(name, value, raw, ':');

				if (name[0] && isIP(name) && inet_addr(name))
				{
					newNetwork.httpConn.loaded = true;
					newNetwork.httpConn.server = inet_addr(name);
				}
				else
				{
					newNetwork.httpConn.loaded = false;
					sprintf(logBuff, "Warning: Section [HTTP_INTERFACE], Invalid IP Address %s, ignored", name);
					logDHCPMess(logBuff, 1);
				}

				if (value[0])
				{
					if (atoi(value))
						newNetwork.httpConn.port = atoi(value);
					else
					{
						newNetwork.httpConn.loaded = false;
						sprintf(logBuff, "Warning: Section [HTTP_INTERFACE], Invalid port %s, ignored", value);
						logDHCPMess(logBuff, 1);
					}
				}

				if (newNetwork.httpConn.server != inet_addr("127.0.0.1") && !findServer(newNetwork.allServers, newNetwork.httpConn.server))
				{
					bindfailed = true;
					newNetwork.httpConn.loaded = false;
					sprintf(logBuff, "Warning: Section [HTTP_INTERFACE], %s not available, ignored", raw);
					logDHCPMess(logBuff, 1);
				}
			}
		}

		if (network.httpConn.ready && network.httpConn.server == newNetwork.httpConn.server)
		{
			memcpy(&(newNetwork.httpConn), &(network.httpConn), sizeof(ConnType));

			if (newNetwork.httpConn.sock > newNetwork.maxFD)
				newNetwork.maxFD = newNetwork.httpConn.sock;

			network.httpConn.ready = false;
		}
		else
		{
			newNetwork.httpConn.sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);

			if (newNetwork.httpConn.sock == INVALID_SOCKET)
			{
				bindfailed = true;
				sprintf(logBuff, "Failed to Create Socket");
				logDHCPMess(logBuff, 1);
			}
			else
			{
				//printf("Socket %u\n", newNetwork.httpConn.sock);

				newNetwork.httpConn.addr.sin_family = AF_INET;
				newNetwork.httpConn.addr.sin_addr.s_addr = newNetwork.httpConn.server;
				newNetwork.httpConn.addr.sin_port = htons(newNetwork.httpConn.port);

				int nRet = bind(newNetwork.httpConn.sock,
								(sockaddr*)&newNetwork.httpConn.addr,
								sizeof(struct sockaddr_in));

				if (nRet == SOCKET_ERROR)
				{
					bindfailed = true;
					sprintf(logBuff, "Http Interface %s TCP Port %u not available", IP2String(tempbuff, newNetwork.httpConn.server), newNetwork.httpConn.port);
					logDHCPMess(logBuff, 1);
					closesocket(newNetwork.httpConn.sock);
				}
				else
				{
					nRet = listen(newNetwork.httpConn.sock, SOMAXCONN);

					if (nRet == SOCKET_ERROR)
					{
						bindfailed = true;
						sprintf(logBuff, "%s TCP Port %u Error on Listen", IP2String(tempbuff, newNetwork.httpConn.server), newNetwork.httpConn.port);
						logDHCPMess(logBuff, 1);
						closesocket(newNetwork.httpConn.sock);
					}
					else
					{
						newNetwork.httpConn.loaded = true;
						newNetwork.httpConn.ready = true;

						if (newNetwork.httpConn.sock > newNetwork.maxFD)
							newNetwork.maxFD = newNetwork.httpConn.sock;
					}
				}
			}
		}

		newNetwork.maxFD++;

		for (MYBYTE m = 0; m < MAX_SERVERS && newNetwork.allServers[m]; m++)
			lockIP(newNetwork.allServers[m]);

		if (bindfailed)
			cfig.failureCount++;
		else
			cfig.failureCount = 0;

		closeConn();
		memcpy(&network, &newNetwork, sizeof(data1));

		//printf("%i %i %i\n", network.dhcpConn[0].ready, network.dnsUdpConn[0].ready, network.dnsTcpConn[0].ready);

		if (!network.dhcpConn[0].ready)
		{
			sprintf(logBuff, "No Static Interface ready, Waiting...");
			logDHCPMess(logBuff, 1);
			continue;
		}

		if (network.httpConn.ready)
		{
			sprintf(logBuff, "Lease Status URL: http://%s:%u", IP2String(tempbuff, network.httpConn.server), network.httpConn.port);
			logDHCPMess(logBuff, 1);
		}

		for (int i = 0; i < MAX_SERVERS && network.staticServers[i]; i++)
		{
			for (MYBYTE j = 0; j < MAX_SERVERS; j++)
			{
				if (network.dhcpConn[j].server == network.staticServers[i])
				{
					sprintf(logBuff, "Listening On: %s", IP2String(tempbuff, network.staticServers[i]));
					logDHCPMess(logBuff, 1);
					break;
				}
			}
		}

		network.ready = true;

	} while (detectChange());

	pthread_exit(NULL);
}

bool detectChange()
{
	//printf("Entering detectchange\n");

    while (kRunning)
    {
		sleep(20);
		//printf("Checking Networks (failue count=%u, failue cycle=%u)..\n", cfig.failureCount, cfig.failureCycle);

		if (!cfig.ifspecified)
			getInterfaces(&newNetwork);

		if (memcmp(cfig.oldservers, newNetwork.staticServers, (MAX_SERVERS * sizeof(MYDWORD))))
		{
			memcpy(cfig.oldservers, newNetwork.staticServers, (MAX_SERVERS * sizeof(MYDWORD)));
			sprintf(logBuff, "Network changed, re-detecting Listening Interfaces..");
			logDHCPMess(logBuff, 1);
			break;
		}
		else if (cfig.failureCount)
		{
			cfig.failureCycle++;

			if (kRunning && cfig.failureCycle == (MYDWORD)pow(2, cfig.failureCount))
			{
				sprintf(logBuff, "Retrying failed Listening Interfaces..");
				logDHCPMess(logBuff, 1);
				break;
			}
		}
		else
			cfig.failureCycle = 0;
	}

	network.ready = false;

	while (kRunning && network.busy)
		sleep(1);

	//printf("Returning from detectchange\n");

	if (kRunning)
		return true;
	else
		return false;
}

void getInterfaces(data1 *network)
{
	memset(network, 0, sizeof(data1));

	Ifc.ifc_len = sizeof(IfcBuf);
	Ifc.ifc_buf = (char*)IfcBuf;

	if (ioctl(cfig.fixedSocket, SIOCGIFCONF, &Ifc) >= 0)
	{

		MYDWORD addr, mask;
		short flags;
		struct ifreq pIfr;
		MYBYTE numInterfaces = Ifc.ifc_len / sizeof(ifreq);

		for (MYBYTE i = 0 ; i < numInterfaces; i++)
		{
			memcpy(&pIfr, &(IfcBuf[i]), sizeof(ifreq));

			if (!ioctl(cfig.fixedSocket, SIOCGIFADDR, &pIfr))
				addr = ((struct sockaddr_in*)&pIfr.ifr_addr)->sin_addr.s_addr;
			else
				addr = 0;

			if (!ioctl(cfig.fixedSocket, SIOCGIFNETMASK, &pIfr))
				mask = ((struct sockaddr_in*)&pIfr.ifr_addr)->sin_addr.s_addr;
			else
				mask = 0;

			if (!ioctl(cfig.fixedSocket, SIOCGIFFLAGS, &pIfr))
				flags = pIfr.ifr_flags;
			else
				flags = 0;


			if (addr && !(flags & IFF_LOOPBACK))
				addServer(network->allServers, addr);

			if (addr && mask && !(flags & IFF_POINTOPOINT) && !(flags & IFF_LOOPBACK) && !(flags & IFF_DYNAMIC))
			{
				if ((flags & IFF_RUNNING) && (flags & IFF_UP))
				{
					for (MYBYTE m = 0; m < MAX_SERVERS; m++)
					{
						if (network->staticServers[m] == addr)
							break;
						else if (!network->staticServers[m])
						{
							network->staticServers[m] = addr;
							network->staticMasks[m] = mask;
							break;
						}
					}
				}
			}
		}
    }
}

void *updateStateFile(void *lparam)
{
	data8 *dhcpData = (data8*)lparam;

	pthread_mutex_lock( &mutStateFile );

	if (dhcpData->dhcpInd)
	{
		FILE *f = fopen(leaFile, "rb+");
		if (f)
		{
			if (fseek(f, (dhcpData->dhcpInd - 1)*sizeof(data8), SEEK_SET) >= 0)
			{
				fwrite(dhcpData, sizeof(data8), 1, f);
			}
			fclose(f);
		}
	}
	else
	{
		dhcpData->dhcpInd = cfig.dhcpInd;

		FILE *f = fopen(leaFile, "ab");
		if (f)
		{
			fwrite(dhcpData, sizeof(data8), 1, f);
			fclose(f);
		}
	}

	free(dhcpData);
	pthread_mutex_unlock( &mutStateFile );
	pthread_exit(NULL);
}

MYDWORD inet_addr(char* str)
{
    if (str == NULL)
        return INADDR_ANY;
    else
    	return inet_addr(str);
}

MYWORD gdmess(data9 *req, MYBYTE sockInd)
{
    memset(req, 0, sizeof(data9));
    errno = 0;

    if (sockInd == 255)
    {
        int msgflags = 0;
        req->iov[0].iov_base = req->raw;
        req->iov[0].iov_len = sizeof(dhcp_packet);
        req->msg.msg_iov = req->iov;
        req->msg.msg_iovlen = 1;
        req->msg.msg_name = &req->remote;
        req->msg.msg_namelen = sizeof(sockaddr_in);
        req->msg.msg_control = &req->msgcontrol;
        req->msg.msg_controllen = sizeof(msg_control);
        req->msg.msg_flags = msgflags;

        int flags = 0;
        req->bytes = recvmsg(network.dhcpListener.sock, &req->msg, flags);

        if (errno || req->bytes <= 0)
            return 0;

        //printf("%u\n", req->msg.msg_controllen);
        //msgcontrol = (msg_control*)msg.msg_control;

        //struct in_addr local_addr;
        //struct in_addr rem_addr;

        //local_addr = msgcontrol->pktinfo.ipi_addr;
        //rem_addr = msgcontrol->pktinfo.ipi_spec_dst;

        //printf("IF = %u\n", req->msgcontrol.pktinfo.ipi_ifindex);
        //printf("LADDR = %s\n", inet_ntoa(req->msgcontrol.pktinfo.ipi_addr));
        //printf("RADDR = %s\n", inet_ntoa(req->msgcontrol.   pktinfo.ipi_spec_dst));


        MYDWORD addr = req->msgcontrol.pktinfo.ipi_spec_dst.s_addr;

        //printf("%s\n",IP2String(tempbuff, addr));

        if (!addr)
            return 0;

        for (int i = 0; ; i++)
        {
            if (i == MAX_SERVERS || !network.dhcpConn[i].server)
                return 0;
            else if (addr == network.dhcpConn[i].server)
            {
                req->sockInd = i;
                break;
            }
        }
    }
    else
    {
        req->sockInd = sockInd;
        req->sockLen = sizeof(req->remote);
        errno = 0;
        req->bytes = recvfrom(network.dhcpConn[sockInd].sock,
                              req->raw,
                              sizeof(dhcp_packet),
                              0,
                              (sockaddr*)&req->remote,
                              &req->sockLen);

        if (errno || req->bytes <= 0)
            return 0;
    }

	if (req->dhcpp.header.bp_op != BOOTP_REQUEST)
		return 0;

	hex2String(req->chaddr, req->dhcpp.header.bp_chaddr, req->dhcpp.header.bp_hlen);

	data3 *op;
	MYBYTE *raw = req->dhcpp.vend_data;
	MYBYTE *rawEnd = raw + (req->bytes - sizeof(dhcp_header));

	for (; raw < rawEnd && *raw != DHCP_OPTION_END;)
	{
		op = (data3*)raw;
		//printf("OpCode=%u,MessType=%u\n", op->opt_code, op->value[0]);

		switch (op->opt_code)
		{
			case DHCP_OPTION_PAD:
				raw++;
				continue;

			case DHCP_OPTION_PARAMREQLIST:
				for (int ix = 0; ix < op->size; ix++)
					req->paramreqlist[op->value[ix]] = 1;
				break;

			case DHCP_OPTION_MESSAGETYPE:
				req->req_type = op->value[0];
				break;

			case DHCP_OPTION_SERVERID:
				req->server = fIP(op->value);
				break;

			case DHCP_OPTION_IPADDRLEASE:
				req->lease = fULong(op->value);
				break;

			case DHCP_OPTION_MAXDHCPMSGSIZE:
				req->messsize = fUShort(op->value);
				break;

			case DHCP_OPTION_REQUESTEDIPADDR:
				req->reqIP = fIP(op->value);
				break;

			case DHCP_OPTION_HOSTNAME:
				if (op->size && strcasecmp((char*)op->value, "(none)") && strcasecmp((char*)op->value, cfig.servername))
				{
					memcpy(req->hostname, op->value, op->size);
					req->hostname[op->size] = 0;

					if (char *ptr = strchr(req->hostname, '.'))
						*ptr = 0;
				}
				break;

			case DHCP_OPTION_VENDORCLASSID:
				memcpy(req->vendClass, op->value, op->size);
				req->vendClassSize = op->size;
				break;

			case DHCP_OPTION_USERCLASS:
				memcpy(req->userClass, op->value, op->size);
				req->userClassSize = op->size;
				break;

			case DHCP_OPTION_RELAYAGENTINFO:
				memcpy(&req->agentOption, op, op->size + 2);
				break;

			case DHCP_OPTION_CLIENTID:
				memcpy(&req->clientId, op, op->size + 2);
				break;

			case DHCP_OPTION_DNS:
				req->dns = fIP(op->value);
				break;
		}
		raw += 2;
		raw += op->size;
	}

	if (!req->messsize)
	{
		if (req->req_type == DHCP_MESS_NONE)
			req->messsize = req->bytes;
		else
			req->messsize = sizeof(dhcp_packet);
	}

	if (cfig.dhcpLogLevel >= 2)
	{
		if (req->req_type == DHCP_MESS_NONE)
		{
			if (req->dhcpp.header.bp_giaddr)
				sprintf(logBuff, "BOOTP request for %s (%s) from RelayAgent %s received", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_giaddr));
			else
				sprintf(logBuff, "BOOTP request for %s (%s) from interface %s received", req->chaddr, req->hostname, IP2String(tempbuff, network.dhcpConn[req->sockInd].server));

			logDHCPMess(logBuff, 2);
		}
		else if (req->req_type == DHCP_MESS_DISCOVER)
		{
			if (req->dhcpp.header.bp_giaddr)
				sprintf(logBuff, "DHCP discover for %s (%s) from RelayAgent %s received", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_giaddr));
			else
				sprintf(logBuff, "DHCP discover for %s (%s) from interface %s received", req->chaddr, req->hostname, IP2String(tempbuff, network.dhcpConn[req->sockInd].server));

			logDHCPMess(logBuff, 2);
		}
		else if (req->req_type == DHCP_MESS_REQUEST)
		{
			if (req->dhcpp.header.bp_giaddr)
				sprintf(logBuff, "DHCP request for %s (%s) from RelayAgent %s received", req->chaddr, req->hostname, IP2String(tempbuff, req->dhcpp.header.bp_giaddr));
			else
				sprintf(logBuff, "DHCP request for %s (%s) from interface %s received", req->chaddr, req->hostname, IP2String(tempbuff, network.dhcpConn[req->sockInd].server));

			logDHCPMess(logBuff, 2);
		}
	}

	req->vp = req->dhcpp.vend_data;
	memset(req->vp, 0, sizeof(dhcp_packet) - sizeof(dhcp_header));
	//printf("end bytes=%u\n", req->bytes);
	return 1;
}

void logDHCPMess(char *logBuff, MYBYTE dhcpLogLevel)
{
//	pthread_mutex_lock( &mutLogFile );

    if (verbatim)
        printf("%s\n", logBuff);
	else if (cfig.logfile && dhcpLogLevel <= cfig.dhcpLogLevel)
	{
		tm *ttm = localtime(&t);

		if (ttm->tm_yday != loggingDay)
		{
			loggingDay = ttm->tm_yday;
			strftime(extbuff, sizeof(extbuff), logFile, ttm);

			if (strcasecmp(currentLogFile, extbuff))
			{
				strcpy(currentLogFile, extbuff);
				fprintf(cfig.logfile, "Logging Continued on file %s\n", currentLogFile);
				fclose(cfig.logfile);
				cfig.logfile = fopen(currentLogFile, "at");

				if (cfig.logfile)
					fprintf(cfig.logfile, "%s\n\n", sVersion);
				else
				{
					sprintf(tempbuff, "Warning: faled to open log file %s", currentLogFile);
					syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), tempbuff);

					if (dhcpLogLevel <= 1)
						syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), logBuff);

					return;
				}
			}
		}

		strftime(extbuff, sizeof(extbuff), "%d-%b-%y %X", ttm);
		fprintf(cfig.logfile, "[%s] %s\n", extbuff, logBuff);
		fflush(cfig.logfile);
	}
    else if (dhcpLogLevel <= 1)
        syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_CRIT), logBuff);

//	pthread_mutex_unlock( &mutLogFile );
}
